#include "Util-pkg.h"
#include "P3d-pkg.h"
#include "Localpath-pkg.h"
#include "Graphic-pkg.h"
#ifdef LIGHT_PLANNER
#include "ManipulationArmData.hpp"
#endif
#include "p3d_sample_proto.h"

#define DEBUG_LEVEL 0
#define PI 3.14159265358979323846
#define EPSI 1e-4
#define NB_IT_MAX 1e3

#define EFLINVEL 0.5
#define JNTMAXANGVEL 3
#define EF_JNT 8

typedef enum {
    A1,
    A2,
    B,
    C1,
    C2,
    D,
    E1,
    E2,
    G,
    H1,
    H2,
    F
} durations_index;

typedef enum {
    A,
    C,
    E,
    H
} signes_index;

void print_int_ac(p3d_quadspline_data *quadspline_specificPt,int comp){
    int k;
    printf("int_ac(%d) --[%d]-------------------------------------------\n",comp,quadspline_specificPt->cases_abc[0][comp]);
    for(k=0;k<=quadspline_specificPt->cases_abc[0][comp];k++){
        printf("%d ",quadspline_specificPt->cases_abc[k][comp]);
    }
    printf("\n");
    for(k=0;k<=quadspline_specificPt->cases_abc[0][comp];k++){
        printf("%f ",quadspline_specificPt->int_v_abc[k][comp]);
    }
    printf("\n");
    for(k=0;k<=quadspline_specificPt->cases_abc[0][comp];k++){
        printf("%f ",quadspline_specificPt->int_a_abc[k][comp]);
    }
    printf("\n");
    printf("----------------------------------------------------------\n");
}

void print_int_eh(p3d_quadspline_data *quadspline_specificPt,int comp){
    int k;
    printf("int_eh(%d) --[%d]-------------------------------------------\n",comp,quadspline_specificPt->cases_egh[0][comp]);
    for(k=0;k<=quadspline_specificPt->cases_egh[0][comp];k++){
        printf("%d ",quadspline_specificPt->cases_egh[k][comp]);
    }
    printf("\n");
    for(k=0;k<=quadspline_specificPt->cases_egh[0][comp];k++){
        printf("%f ",quadspline_specificPt->int_v_egh[k][comp]);
    }
    printf("\n");
    for(k=0;k<=quadspline_specificPt->cases_egh[0][comp];k++){
        printf("%f ",quadspline_specificPt->int_a_egh[k][comp]);
    }
    printf("\n");
    printf("----------------------------------------------------------\n");
}

double sign(double d){
    if(d<0) return -1.;
    if(d>0) return  1.;
    return 0.;
}

double * poly_root_2(double a,double b,double c){
    double d;
    double *sol;
    if(fabs(a)<EPS10){
        sol=(double*)malloc(2*sizeof(double));
        if(!sol) return sol;
        sol[0]=1;
        if(fabs(b)<EPS10){
            sol[1]=-c;
            return sol;
        }
        sol[1]=-c/b;
        return sol;
    }
    d=b*b-4*a*c;
    if(fabs(d)<EPS10){
        sol=(double*)malloc(2*sizeof(double));
        if(!sol) return sol;
        sol[0]=1;
        sol[1]=-b/(2*a);
        return sol;
    }
    if(d>0){
        sol=(double*)malloc(3*sizeof(double));
        if(!sol) return sol;
        sol[0]=2;
        sol[1]=(-b-sqrt(d))/(2*a);
        sol[2]=(-b+sqrt(d))/(2*a);
        return sol;
    }
    sol=(double*)malloc(1*sizeof(double));
    if(!sol) return sol;
    sol[0]=0;
    return sol;
}

double * cardan(double p,double q){
    double *sol;
    double u,v;
    double delta=-(4*p*p*p+27*q*q);
    int k;
    if(fabs(delta)<EPS10){
        if(fabs(p)<EPS10 && fabs(q)<EPS10){
            sol=(double*)malloc(2*sizeof(double));
            sol[0]=1;
            sol[1]=0.;
            return sol;
        }
        else{
            sol=(double*)malloc(3*sizeof(double));
            sol[0]=2;
            sol[1]=3*q/p;
            sol[2]=-3*q/(2*p);
            return sol;
        }
    }
    if(delta<0){
        v=(sqrt(-delta/27.))/2.;
        u=-q/2+v;
        u=sign(u)*pow(fabs(u),0.333333333333);
        v=-q/2-v;
        v=sign(v)*pow(fabs(v),0.333333333333);
        sol=(double*)malloc(2*sizeof(double));
        if(!sol) return sol;
        sol[0]=1;
        sol[1]=u+v;
        return sol;
    }
    u=0.333333333333*acos((-q/2)*sqrt(27/(-p*p*p)));
    v=2*sqrt(-p/3);
    sol=(double*)malloc(4*sizeof(double));
    if(!sol) return sol;
    sol[0]=3;
    for(k=0;k<3;k++){
        sol[k+1]=v*cos(u+2*k*M_PI/3);
    }
    return sol;
}

double * poly_root_3(double a,double b,double c,double d){
    double aa,bb,cc,p,q;
    double *sol;
    int k;
    if(fabs(a)<EPS10) return poly_root_2(b,c,d);
    aa=b/a;
    bb=c/a;
    cc=d/a;
    p=bb-aa*aa/3.;
    q=2*aa*aa*aa/27.-aa*bb/3.+cc;
    sol=cardan(p,q);
    for(k=1;k<=(int)sol[0];k++){
        sol[k]=sol[k]-aa/3;
    }
    return sol;
}

double * ferrari(double p,double q,double r){
    double *sol,*soltmp1,*soltmp2;
    double delta,y0,a0,b0;
    int k,nb_sol,i;
    if(fabs(q)<EPS10){
        delta=p*p-4.*r;
        soltmp1=poly_root_2(1.,p,r);
        nb_sol=0;
        for(k=1;k<=(int)soltmp1[0];k++){
            if(soltmp1[k]>=0) nb_sol++;
        }
        sol=(double*)malloc((nb_sol+1)*sizeof(double));
        sol[0]=nb_sol;
        i=1;
        for(k=1;k<=(int)soltmp1[0];k++){
            if(soltmp1[k]>=0){
                sol[i]=-sqrt(soltmp1[k]);
                sol[i+1]=sqrt(soltmp1[k]);
                i=i+2;
            }
        }
        free(soltmp1);
        return sol;
    }
    soltmp1=poly_root_3(1.,-p/2.,-r,p*r/2.-q*q/8.);
    y0=soltmp1[1];
    free(soltmp1);
    if(2*y0-p<0){
        sol=poly_root_2(1.,0.,y0);
        return sol;
    }
    a0=sqrt(2*y0-p);
    if(fabs(2*y0-p)<EPS10){
        if(y0*y0-r>0){
            b0=sqrt(y0*y0-r);
        }
        else{
            b0=0;
        }
    }
    else{
        b0=-q/(2*a0);
    }
    soltmp1=poly_root_2(1.,a0,y0+b0);
    soltmp2=poly_root_2(1.,-a0,y0-b0);
    nb_sol=(int)soltmp1[0]+(int)soltmp2[0];
    sol=(double*)malloc((nb_sol+1)*sizeof(double));
    sol[0]=nb_sol;
    for(k=1;k<=(int)soltmp1[0];k++){
        sol[k]=soltmp1[k];
    }
    for(k=1;k<=(int)soltmp2[0];k++){
        sol[k+(int)soltmp1[0]]=soltmp2[k];
    }
    free(soltmp1);
    free(soltmp2);
    return sol;
}

double * poly_root_4(double a,double b,double c,double d,double e){
    double *sol;
    double aa,bb,cc,dd,p,q,r;
    int k;
    if(fabs(a)<EPS10) return poly_root_3(b,c,d,e);
    aa=b/a;
    bb=c/a;
    cc=d/a;
    dd=e/a;
    p=bb-3.*aa*aa/8.;
    q=cc-aa*bb/2.+aa*aa*aa/8.;
    r=dd-aa*cc/4.+bb*aa*aa/16.-3.*aa*aa*aa*aa/256.;
    sol=ferrari(p,q,r);
    for(k=1;k<=(int)sol[0];k++){
        sol[k]=sol[k]-aa/4.;
    }
    return sol;
}

int attitude(double Tx,double Ty,double Tz,double *rx,double *ry){
    double norm_T=sqrt(SQR(Tx)+SQR(Ty)+SQR(Tz));
    if(norm_T<EPS10){
        *rx=0;
        *ry=0;
        return 0;
    }
    *ry=asin(Tx/norm_T);
    if(fabs(Ty)<EPS10 && fabs(Tz)<EPS10){
        *rx=0;
        return 1;
    }
    if(Tz<0){
        *rx=PI+asin(Ty/sqrt(SQR(Ty)+SQR(Tz)));
    }
    else{
        *rx=-asin(Ty/sqrt(SQR(Ty)+SQR(Tz)));
    }
    return 1;
}

void set_attitude(configPt q,lm_quadspline_str *quadspline_paramPt){
    double Tx,Ty,Tz,m,f,g,vx,vy,vz,ax,ay,az,rx,ry;
    m=quadspline_paramPt->mass;
    g=quadspline_paramPt->gravity;
    f=quadspline_paramPt->friction;
    vx=q[quadspline_paramPt->vx_dof];
    vy=q[quadspline_paramPt->vy_dof];
    vz=q[quadspline_paramPt->vz_dof];
    ax=q[quadspline_paramPt->ax_dof];
    ay=q[quadspline_paramPt->ay_dof];
    az=q[quadspline_paramPt->az_dof];
    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;
    attitude(Tx,Ty,Tz,&rx,&ry);
    q[quadspline_paramPt->rx_dof]=rx;
    q[quadspline_paramPt->ry_dof]=ry;
}

int case_abc(double aB,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){
    int sA,sC,cond_A,cond_C,case_ABC;
    double a0,jmax,smax;
    if(comp<3){
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        a0=0;
    }
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    sA=(int)sign(aB-a0);
    sC=-(int)sign(aB);
    cond_A=(fabs(a0-aB)>SQR(jmax)/smax);
    cond_C=(fabs(aB)>SQR(jmax)/smax);
    case_ABC=1000*(sA+1)+100*(sC+1)+10*cond_A+cond_C;
    return case_ABC;
}

int case_egh(double aG,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){
    int sH,sE,cond_H,cond_E,case_EGH;
    double aF,jmax,smax;
    if(comp<3){
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        aF=0;
    }
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    sH=(int)sign(aF-aG);
    sE=(int)sign(aG);
    cond_H=(fabs(aG-aF)>SQR(jmax)/smax);
    cond_E=(fabs(aG)>SQR(jmax)/smax);
    case_EGH=1000*(sH+1)+100*(sE+1)+10*cond_H+cond_E;
    return case_EGH;
}

void durations_and_signs_ac(double aB,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){
    double a0,jmax,smax;
    if(comp<3){
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        a0=0;
    }
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    quadspline_specificPt->signs[A][comp]=sign(aB-a0);
    quadspline_specificPt->signs[C][comp]=-sign(aB);
    if(fabs(aB-a0)>SQR(jmax)/smax){
        quadspline_specificPt->durations[A1][comp]=jmax/smax;
        quadspline_specificPt->durations[A2][comp]=fabs(aB-a0)/jmax-jmax/smax;
    }
    else{
        quadspline_specificPt->durations[A1][comp]=sqrt(fabs(a0-aB)/smax);
        quadspline_specificPt->durations[A2][comp]=0;
    }
    if(fabs(aB)>SQR(jmax)/smax){
        quadspline_specificPt->durations[C1][comp]=jmax/smax;
        quadspline_specificPt->durations[C2][comp]=fabs(aB)/jmax-jmax/smax;
    }
    else{
        quadspline_specificPt->durations[C1][comp]=sqrt(fabs(aB)/smax);
        quadspline_specificPt->durations[C2][comp]=0;
    }
}

void durations_and_signs_eh(double aG,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){
    double aF,jmax,smax;
    if(comp<3){
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        aF=0;
    }
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    quadspline_specificPt->signs[H][comp]=sign(aF-aG);
    quadspline_specificPt->signs[E][comp]=sign(aG);
    if(fabs(aF-aG)>SQR(jmax)/smax){
        quadspline_specificPt->durations[H1][comp]=jmax/smax;
        quadspline_specificPt->durations[H2][comp]=fabs(aF-aG)/jmax-jmax/smax;
    }
    else{
        quadspline_specificPt->durations[H1][comp]=sqrt(fabs(aF-aG)/smax);
        quadspline_specificPt->durations[H2][comp]=0;
    }
    if(fabs(aG)>SQR(jmax)/smax){
        quadspline_specificPt->durations[E1][comp]=jmax/smax;
        quadspline_specificPt->durations[E2][comp]=fabs(aG)/jmax-jmax/smax;
    }
    else{
        quadspline_specificPt->durations[E1][comp]=sqrt(fabs(aG)/smax);
        quadspline_specificPt->durations[E2][comp]=0;
    }
}

double x_c(double aB,double new_tB,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,int use_old){
    double x0,v0,a0,smax;
    double tA1,tA2,tB,tC1,tC2,sA,sC;
    double tA1_2,tA2_2,tB_2,tC1_2,tC2_2;
    double tA1_3,tA2_3,tC1_3,tC2_3;
    double tA1_4,tC1_4;
    if(!use_old){
        quadspline_specificPt->durations[B][comp]=new_tB;
        durations_and_signs_ac(aB,quadspline_paramPt,quadspline_specificPt,comp);
    }
    tA1=quadspline_specificPt->durations[A1][comp];
    tA2=quadspline_specificPt->durations[A2][comp];
    tB=quadspline_specificPt->durations[B][comp];
    tC1=quadspline_specificPt->durations[C1][comp];
    tC2=quadspline_specificPt->durations[C2][comp];
    sA=quadspline_specificPt->signs[A][comp];
    sC=quadspline_specificPt->signs[C][comp];
    if(comp<3){
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+comp];
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+comp];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
        v0=0;
        a0=0;
    }
    smax=quadspline_paramPt->smax;
    tA1_2=tA1*tA1;
    tA1_3=tA1_2*tA1;
    tA1_4=tA1_3*tA1;
    tA2_2=tA2*tA2;
    tA2_3=tA2_2*tA2;
    tB_2=tB*tB;
    tC1_2=tC1*tC1;
    tC1_3=tC1_2*tC1;
    tC1_4=tC1_3*tC1;
    tC2_2=tC2*tC2;
    tC2_3=tC2_2*tC2;
    return (7*sA*smax*tA1_4)/12+(7*sA*smax*tA1_3*tA2)/6+sA*smax*tA1_3*tB+2*sA*smax*tA1_3*tC1+sA*smax*tA1_3*tC2+(3*sA*smax*tA1_2*tA2_2)/4+(3*sA*smax*tA1_2*tA2*tB)/2+3*sA*smax*tA1_2*tA2*tC1+(3*sA*smax*tA1_2*tA2*tC2)/2+(sA*smax*tA1_2*tB_2)/2+2*sA*smax*tA1_2*tB*tC1+sA*smax*tA1_2*tB*tC2+2*sA*smax*tA1_2*tC1_2+2*sA*smax*tA1_2*tC1*tC2+(sA*smax*tA1_2*tC2_2)/2+2*a0*tA1_2+(sA*smax*tA1*tA2_3)/6+(sA*smax*tA1*tA2_2*tB)/2+sA*smax*tA1*tA2_2*tC1+(sA*smax*tA1*tA2_2*tC2)/2+(sA*smax*tA1*tA2*tB_2)/2+2*sA*smax*tA1*tA2*tB*tC1+sA*smax*tA1*tA2*tB*tC2+2*sA*smax*tA1*tA2*tC1_2+2*sA*smax*tA1*tA2*tC1*tC2+(sA*smax*tA1*tA2*tC2_2)/2+2*a0*tA1*tA2+2*a0*tA1*tB+4*a0*tA1*tC1+2*a0*tA1*tC2+2*v0*tA1+(a0*tA2_2)/2+a0*tA2*tB+2*a0*tA2*tC1+a0*tA2*tC2+v0*tA2+(a0*tB_2)/2+2*a0*tB*tC1+a0*tB*tC2+v0*tB+(7*sC*smax*tC1_4)/12+(7*sC*smax*tC1_3*tC2)/6+(3*sC*smax*tC1_2*tC2_2)/4+2*a0*tC1_2+(sC*smax*tC1*tC2_3)/6+2*a0*tC1*tC2+2*v0*tC1+(a0*tC2_2)/2+v0*tC2+x0;
}

double x_e(double aG,double new_tG,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,int use_old){
    double xF,vF,aF,smax;
    double tE1,tE2,tG,tH1,tH2,sE,sH;
    double tE1_2,tE2_2,tG_2,tH1_2,tH2_2;
    double tE1_3,tE2_3,tH1_3,tH2_3;
    double tE1_4,tH1_4;
    if(!use_old){
        quadspline_specificPt->durations[G][comp]=new_tG;
        durations_and_signs_eh(aG,quadspline_paramPt,quadspline_specificPt,comp);
    }
    tE1=quadspline_specificPt->durations[E1][comp];
    tE2=quadspline_specificPt->durations[E2][comp];
    tG=quadspline_specificPt->durations[G][comp];
    tH1=quadspline_specificPt->durations[H1][comp];
    tH2=quadspline_specificPt->durations[H2][comp];
    sE=quadspline_specificPt->signs[E][comp];
    sH=quadspline_specificPt->signs[H][comp];
    if(comp<3){
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+comp];
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+comp];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
        vF=0;
        aF=0;
    }
    smax=quadspline_paramPt->smax;
    tE1_2=tE1*tE1;
    tE1_3=tE1_2*tE1;
    tE1_4=tE1_3*tE1;
    tE2_2=tE2*tE2;
    tE2_3=tE2_2*tE2;
    tG_2=tG*tG;
    tH1_2=tH1*tH1;
    tH1_3=tH1_2*tH1;
    tH1_4=tH1_3*tH1;
    tH2_2=tH2*tH2;
    tH2_3=tH2_2*tH2;
    return -(7*sE*smax*tE1_4)/12-(7*sE*smax*tE1_3*tE2)/6-(3*sE*smax*tE1_2*tE2_2)/4-2*sH*smax*tE1_2*tH1_2-2*sH*smax*tE1_2*tH1*tH2+2*aF*tE1_2-(sE*smax*tE1*tE2_3)/6-2*sH*smax*tE1*tE2*tH1_2-2*sH*smax*tE1*tE2*tH1*tH2+2*aF*tE1*tE2-2*sH*smax*tE1*tG*tH1_2-2*sH*smax*tE1*tG*tH1*tH2+2*aF*tE1*tG-2*sH*smax*tE1*tH1_3-3*sH*smax*tE1*tH1_2*tH2-sH*smax*tE1*tH1*tH2_2+4*aF*tE1*tH1+2*aF*tE1*tH2-2*vF*tE1-(sH*smax*tE2_2*tH1_2)/2-(sH*smax*tE2_2*tH1*tH2)/2+(aF*tE2_2)/2-sH*smax*tE2*tG*tH1_2-sH*smax*tE2*tG*tH1*tH2+aF*tE2*tG-sH*smax*tE2*tH1_3-(3*sH*smax*tE2*tH1_2*tH2)/2-(sH*smax*tE2*tH1*tH2_2)/2+2*aF*tE2*tH1+aF*tE2*tH2-vF*tE2-(sH*smax*tG_2*tH1_2)/2-(sH*smax*tG_2*tH1*tH2)/2+(aF*tG_2)/2-sH*smax*tG*tH1_3-(3*sH*smax*tG*tH1_2*tH2)/2-(sH*smax*tG*tH1*tH2_2)/2+2*aF*tG*tH1+aF*tG*tH2-vF*tG-(7*sH*smax*tH1_4)/12-(7*sH*smax*tH1_3*tH2)/6-(3*sH*smax*tH1_2*tH2_2)/4+2*aF*tH1_2-(sH*smax*tH1*tH2_3)/6+2*aF*tH1*tH2-2*vF*tH1+(aF*tH2_2)/2-vF*tH2+xF;
}

double v_c(double aB,double new_tB,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,int use_old){
    double v0,a0,smax;
    double tA1,tA2,tB,tC1,tC2,sA,sC;
    double tA1_2,tA2_2,tC1_2,tC2_2;
    double tA1_3,tC1_3;
    if(!use_old){
        quadspline_specificPt->durations[B][comp]=new_tB;
        durations_and_signs_ac(aB,quadspline_paramPt,quadspline_specificPt,comp);
    }
    tA1=quadspline_specificPt->durations[A1][comp];
    tA2=quadspline_specificPt->durations[A2][comp];
    tB=quadspline_specificPt->durations[B][comp];
    tC1=quadspline_specificPt->durations[C1][comp];
    tC2=quadspline_specificPt->durations[C2][comp];
    sA=quadspline_specificPt->signs[A][comp];
    sC=quadspline_specificPt->signs[C][comp];
    if(comp<3){
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+comp];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        v0=0;
        a0=0;
    }
    smax=quadspline_paramPt->smax;
    tA1_2=tA1*tA1;
    tA1_3=tA1_2*tA1;
    tA2_2=tA2*tA2;
    tC1_2=tC1*tC1;
    tC1_3=tC1_2*tC1;
    tC2_2=tC2*tC2;
    return sA*smax*tA1_3+(3*sA*smax*tA1_2*tA2)/2+2*sA*smax*tA1_2*tC1+sA*smax*tA1_2*tC2+sA*smax*tB*tA1_2+(sA*smax*tA1*tA2_2)/2+2*sA*smax*tA1*tA2*tC1+sA*smax*tA1*tA2*tC2+sA*smax*tB*tA1*tA2+2*a0*tA1+a0*tA2+sC*smax*tC1_3+(3*sC*smax*tC1_2*tC2)/2+(sC*smax*tC1*tC2_2)/2+2*a0*tC1+a0*tC2+v0+a0*tB;;
}

double v_e(double aG,double new_tG,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,int use_old){
    double vF,aF,smax;
    double tE1,tE2,tG,tH1,tH2,sE,sH;
    double tE1_2,tE2_2,tH1_2,tH2_2;
    double tE1_3,tH1_3;
    if(!use_old){
        quadspline_specificPt->durations[G][comp]=new_tG;
        durations_and_signs_eh(aG,quadspline_paramPt,quadspline_specificPt,comp);
    }
    tE1=quadspline_specificPt->durations[E1][comp];
    tE2=quadspline_specificPt->durations[E2][comp];
    tG=quadspline_specificPt->durations[G][comp];
    tH1=quadspline_specificPt->durations[H1][comp];
    tH2=quadspline_specificPt->durations[H2][comp];
    sE=quadspline_specificPt->signs[E][comp];
    sH=quadspline_specificPt->signs[H][comp];
    if(comp<3){
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+comp];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        vF=0;
        aF=0;
    }
    smax=quadspline_paramPt->smax;
    tE1_2=tE1*tE1;
    tE1_3=tE1_2*tE1;
    tE2_2=tE2*tE2;
    tH1_2=tH1*tH1;
    tH1_3=tH1_2*tH1;
    tH2_2=tH2*tH2;
    return sE*smax*tE1_3+(3*sE*smax*tE1_2*tE2)/2+(sE*smax*tE1*tE2_2)/2+2*sH*smax*tE1*tH1_2+2*sH*smax*tE1*tH1*tH2-2*aF*tE1+sH*smax*tE2*tH1_2+sH*smax*tE2*tH1*tH2-aF*tE2+sH*smax*tH1_3+(3*sH*smax*tH1_2*tH2)/2+sH*smax*tG*tH1_2+(sH*smax*tH1*tH2_2)/2+sH*smax*tG*tH1*tH2-2*aF*tH1-aF*tH2+vF-aF*tG;
}

void intervals_ac(lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] intervals_ac (comp=%d)\n",comp);
    }

    double v0,a0,amax,jmax,smax,aj,a1,a2;
    int i,k,l,n=1;
    if(comp<3){
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+comp];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        v0=0;
        a0=0;
    }
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    aj=SQR(jmax)/smax;
    double Ac[7]={amax,0,aj,-aj,a0,a0+aj,a0-aj};
    quadspline_specificPt->int_a_abc[0][comp]=-amax;
    for(k=0;k<7;k++){
        if((Ac[k]*a0<=0 || fabs(Ac[k])>=fabs(a0)) && fabs(Ac[k])<=amax){
            l=0;
            while(l<n && Ac[k]>quadspline_specificPt->int_a_abc[l][comp]){
                l++;
            }
            if(l==n || fabs(Ac[k]-quadspline_specificPt->int_a_abc[l][comp])>EPSI){
                for(i=n+1;i>l;i--){
                    quadspline_specificPt->int_a_abc[i][comp]=quadspline_specificPt->int_a_abc[i-1][comp];
                }
                quadspline_specificPt->int_a_abc[l][comp]=Ac[k];
                n++;
            }
        }
    }
    n--;
    quadspline_specificPt->cases_abc[0][comp]=n;
    a1=quadspline_specificPt->int_a_abc[0][comp];
    quadspline_specificPt->int_v_abc[0][comp]=v_c(a1,0,quadspline_paramPt,quadspline_specificPt,comp,0);
    for(k=1;k<=n;k++){
        a1=quadspline_specificPt->int_a_abc[k-1][comp];
        a2=quadspline_specificPt->int_a_abc[k][comp];
        quadspline_specificPt->int_v_abc[k][comp]=v_c(a2,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        quadspline_specificPt->cases_abc[k][comp]=case_abc((a1+a2)/2,quadspline_paramPt,quadspline_specificPt,comp);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] intervals_ac (comp=%d)\n",comp);
    }

}

void intervals_eh(lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] intervals_eh (comp=%d)\n",comp);
    }

    double vF,aF,amax,jmax,smax,aj,a1,a2;
    int i,k,l,n=1;
    if(comp<3){
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+comp];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        vF=0;
        aF=0;
    }
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    smax=quadspline_paramPt->smax;
    aj=SQR(jmax)/smax;
    double Ac[7]={amax,0,aj,-aj,aF,aF+aj,aF-aj};
    quadspline_specificPt->int_a_egh[0][comp]=-amax;
    for(k=0;k<7;k++){
        if((Ac[k]*aF<=0 || fabs(Ac[k])>=fabs(aF)) && fabs(Ac[k])<=amax){
            l=0;
            while(l<n && Ac[k]<quadspline_specificPt->int_a_egh[l][comp]){
                l++;
            }
            if(l==n || fabs(Ac[k]-quadspline_specificPt->int_a_egh[l][comp])>EPSI){
                for(i=n+1;i>l;i--){
                    quadspline_specificPt->int_a_egh[i][comp]=quadspline_specificPt->int_a_egh[i-1][comp];
                }
                quadspline_specificPt->int_a_egh[l][comp]=Ac[k];
                n++;
            }
        }
    }
    n--;
    quadspline_specificPt->cases_egh[0][comp]=n;
    a1=quadspline_specificPt->int_a_egh[0][comp];
    quadspline_specificPt->int_v_egh[0][comp]=v_e(a1,0,quadspline_paramPt,quadspline_specificPt,comp,0);
    for(k=1;k<=n;k++){
        a1=quadspline_specificPt->int_a_egh[k-1][comp];
        a2=quadspline_specificPt->int_a_egh[k][comp];
        quadspline_specificPt->int_v_egh[k][comp]=v_e(a2,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        quadspline_specificPt->cases_egh[k][comp]=case_egh((a1+a2)/2,quadspline_paramPt,quadspline_specificPt,comp);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] intervals_eh (comp=%d)\n",comp);
    }

}

double a_b(double vC,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] a_b (comp=%d) (vC=%f)\n",comp,vC);
    }

    double v0,a0,a0_2,a0_3,amax,jmax,jmax_2,jmax_4,smax,smax_2;
    double a,b,c,d,*sol,dV,dVmin,aB,mina,maxa;
    int nb_int,k,case_ABC,N,index_int;
    if(comp<3){
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+comp];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        v0=0;
        a0=0;
    }
    a0_2=a0*a0;
    a0_3=a0_2*a0;
    nb_int=quadspline_specificPt->cases_abc[0][comp];
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    jmax_2=jmax*jmax;
    jmax_4=jmax_2*jmax_2;
    smax=quadspline_paramPt->smax;
    smax_2=smax*smax;

    if(vC<quadspline_specificPt->int_v_abc[0][comp]){
        quadspline_specificPt->durations[B][comp]=(quadspline_specificPt->int_v_abc[0][comp]-vC)/amax;
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d) (vC<min_v)\n",comp);
        }
        return -amax;
    }

    if(vC>quadspline_specificPt->int_v_abc[nb_int][comp]){
        quadspline_specificPt->durations[B][comp]=(vC-quadspline_specificPt->int_v_abc[nb_int][comp])/amax;
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d) (vC>max_v)\n",comp);
        }
        return amax;
    }

    quadspline_specificPt->durations[B][comp]=0;
    k=1;
    while(k<=nb_int && vC>quadspline_specificPt->int_v_abc[k][comp]){
        k++;
    }
    index_int=k;
    if(index_int>nb_int){
        printf("ERROR in a_b : vC not found\n");
        print_int_ac(quadspline_specificPt,comp);
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d)\n",comp);
        }
        return a0;
    }
    /*
    if(fabs(vC-quadspline_specificPt->int_v_abc[index_int][comp])<EPSI){
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d) (|vC-int_v_abc[%d]|<EPSI)\n",comp,k);
        }
        return quadspline_specificPt->int_a_abc[index_int][comp];
    }
    */
    case_ABC=quadspline_specificPt->cases_abc[index_int][comp];
    switch(case_ABC){
        case 210:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_b (comp=%d) (direct) (case_ABC=%d)\n",comp,case_ABC);
            }
            return -SQR(jmax-sqrt(4*sqrt(smax*(a0*jmax_2+a0_2*smax+2*jmax*smax*(v0-vC)))+jmax_2))/(4*smax);
        case 211:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_b (comp=%d) (direct) (case_ABC=%d)\n",comp,case_ABC);
            }
            return (jmax_2-sqrt(jmax_4+2*a0_2*smax_2+4*jmax*smax_2*(v0-vC)+2*a0*jmax_2*smax))/(2*smax);
        case 2010:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_b (comp=%d) (direct) (case_ABC=%d)\n",comp,case_ABC);
            }
            return SQR(jmax-sqrt(4*sqrt(smax*(a0_2*smax-a0*jmax_2+2*jmax*smax*(vC-v0)))+jmax_2))/(4*smax);
        case 2011:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_b (comp=%d) (direct) (case_ABC=%d)\n",comp,case_ABC);
            }
            return (sqrt(jmax_4+2*a0_2*smax_2+4*jmax*smax_2*(vC-v0)-2*a0*jmax_2*smax)-jmax_2)/(2*smax);
        case 200:
            a=2*sqrt(smax)*(v0-vC)/a0;
            b=-a0;
            c=-4*a0*sqrt(smax)*(v0-vC)/a0;
            d=-(smax*SQR(v0-vC)+a0_3)/a0;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=-SQR(sol[k])+a0;
            }
            break;
        case 201:
            a=2*jmax/sqrt(smax);
            b=(jmax_2-2*a0*smax)/smax;
            c=-4*a0*jmax/sqrt(smax);
            d=-(a0*jmax_2-a0_2*smax+2*jmax*smax*(v0-vC))/smax;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=-SQR(sol[k])+a0;
            }
            break;
        case 2000:
            a=2*sqrt(smax)*(vC-v0)/a0;
            b=-a0;
            c=0;
            d=-(smax*SQR(vC-v0)+a0_3)/a0;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=SQR(sol[k]);
            }
            break;
        case 2001:
            a=2*jmax/sqrt(smax);
            b=(2*a0*smax+jmax_2)/smax;
            c=4*a0*jmax/sqrt(smax);
            d=(a0*jmax_2+a0_2*smax+2*jmax*smax*(v0-vC))/smax;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=SQR(sol[k])+a0;
            }
            break;
        default:
            if(DEBUG_LEVEL>0){
                printf("ERROR in a_b : unknown case : %d\n",case_ABC);
                print_int_ac(quadspline_specificPt,comp);
                printf("ALEX_DEBUG : [END] a_b (comp=%d)\n",comp);
            }
            return quadspline_specificPt->int_a_abc[index_int][comp];
    }
    N=(int)sol[0];
    if(N>0){
        dVmin=fabs(vC-v_c(sol[1],0,quadspline_paramPt,quadspline_specificPt,comp,0));
        aB=sol[1];
        for(k=2;k<=N;k++){
            dV=fabs(vC-v_c(sol[k],0,quadspline_paramPt,quadspline_specificPt,comp,0));
            if(dV<dVmin){
                dVmin=dV;
                aB=sol[k];
            }
        }
    }
    else{
        dV=EPSI+1;
    }
    free(sol);
    if(dV>EPSI){
        mina=quadspline_specificPt->int_a_abc[index_int-1][comp];
        maxa=quadspline_specificPt->int_a_abc[index_int][comp];
        aB=(mina+maxa)/2;
        dV=vC-v_c(aB,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        while(fabs(dV)>EPSI && fabs(mina-maxa)>EPSI){
            if(dV>0){
                mina=aB;
            }
            else{
                maxa=aB;
            }
            aB=(mina+maxa)/2;
            dV=vC-v_c(aB,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        }
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d) (dicho) (case_ABC=%d)\n",comp,case_ABC);

        }
    }
    else{
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_b (comp=%d) (root) (case_ABC=%d)\n",comp,case_ABC);

        }
    }

    return aB;
}

double a_g(double vE,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] a_g (comp=%d) (vE=%f)\n",comp,vE);
    }

    double vF,aF,aF_2,aF_3,amax,jmax,jmax_2,jmax_4,smax,smax_2;
    double a,b,c,d,*sol,dV,dVmin,aG,mina,maxa;
    int nb_int,k,case_EGH,N,index_int;
    if(comp<3){
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+comp];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        vF=0;
        aF=0;
    }
    aF_2=SQR(aF);
    aF_3=aF_2*aF;
    nb_int=quadspline_specificPt->cases_egh[0][comp];
    amax=quadspline_paramPt->amax;
    jmax=quadspline_paramPt->jmax;
    jmax_2=SQR(jmax);
    jmax_4=SQR(jmax_2);
    smax=quadspline_paramPt->smax;
    smax_2=SQR(smax);
    if(vE<quadspline_specificPt->int_v_egh[0][comp]){
        quadspline_specificPt->durations[G][comp]=(quadspline_specificPt->int_v_egh[0][comp]-vE)/amax;
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (vE<min_v)\n",comp);
        }
        return amax;
    }
    if(vE>quadspline_specificPt->int_v_egh[nb_int][comp]){
        quadspline_specificPt->durations[G][comp]=(vE-quadspline_specificPt->int_v_egh[nb_int][comp])/amax;
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (vE>max_v)\n",comp);
        }
        return -amax;
    }
    quadspline_specificPt->durations[G][comp]=0;
    if(fabs(vE-quadspline_specificPt->int_v_egh[0][comp])<EPSI){
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (|vE-min_v|<EPSI)\n",comp);
        }
        return quadspline_specificPt->int_a_egh[0][comp];
    }
    k=1;
    while(k<=nb_int && vE>quadspline_specificPt->int_v_egh[k][comp]){
        k++;
    }
    index_int=k;
    if(index_int>nb_int){
        if(DEBUG_LEVEL>0){
            printf("ERROR in a_g : vE not found : %f\n",vE);
            print_int_eh(quadspline_specificPt,comp);
            printf("ALEX_DEBUG : [END] a_g (comp=%d)\n",comp);
        }
        return aF;
    }
    /*
    if(fabs(vE-quadspline_specificPt->int_v_egh[index_int][comp])<EPSI){
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (|vE-int_v_egh[%d]|)<EPSI\n",comp,k);
        }
        return quadspline_specificPt->int_a_egh[index_int][comp];
    }
    */
    //case_EGH=quadspline_specificPt->cases_egh[index_int][comp];
    case_EGH=case_egh((quadspline_specificPt->int_a_egh[index_int-1][comp]+quadspline_specificPt->int_a_egh[index_int][comp])/2,quadspline_paramPt,quadspline_specificPt,comp);
    switch(case_EGH){
        case 210:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_g (comp=%d) (direct) (case_EGH=%d)\n",comp,case_EGH);
            }
            return SQR(jmax-sqrt(4*sqrt(smax*(aF_2*smax-aF*jmax_2+2*jmax*smax*(vF-vE)))+jmax_2))/(4*smax);
        case 211:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_g (comp=%d) (direct) (case_EGH=%d)\n",comp,case_EGH);
            }
            return (sqrt(jmax_4+2*aF_2*smax_2-2*aF*jmax_2*smax+4*jmax*smax_2*(vF-vE))-jmax_2)/(2*smax);
        case 2010:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_g (comp=%d) (direct) (case_EGH=%d)\n",comp,case_EGH);
            }
            return -SQR(jmax-sqrt(4*sqrt(smax*(aF*jmax_2+aF_2*smax+2*jmax*smax*(vE-vF)))+jmax_2))/(4*smax);
        case 2011:
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_g (comp=%d) (direct) (case_EGH=%d)\n",comp,case_EGH);
            }
            return (jmax_2-sqrt(jmax_4+2*aF_2*smax_2+2*aF*jmax_2*smax+4*jmax*smax_2*(vE-vF)))/(2*smax);
        case 200:
            a=2*sqrt(smax)*(vF-vE)/aF;
            b=-aF;
            c=0;
            d=-(smax*SQR(vE-vF)+aF_3)/aF;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=SQR(sol[k]);
            }
            break;
        case 201:
            a=2*jmax/sqrt(smax);
            b=(2*aF*smax+jmax_2)/smax;
            c=4*aF*jmax/sqrt(smax);
            d=(aF*jmax_2+aF_2*smax+2*jmax*smax*(vE-vF))/smax;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=SQR(sol[k])+aF;
            }
            break;
        case 2000:
            a=2*sqrt(smax)*(vE-vF)/aF;
            b=-aF;
            c=4*aF*sqrt(smax)*(vF-vE)/aF;
            d=-(smax*SQR(vE-vF)+aF_3)/aF;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=-SQR(sol[k])+aF;
            }
            break;
        case 2001:
            a=2*jmax/sqrt(smax);
            b=(jmax_2-2*aF*smax)/smax;
            c=-4*aF*jmax/sqrt(smax);
            d=(aF_2*smax-aF*jmax_2+2*jmax*smax*(vF-vE))/smax;
            sol=poly_root_4(1,a,b,c,d);
            for(k=1;k<=(int)sol[0];k++){
                sol[k]=-SQR(sol[k])+aF;
            }
            break;
        default:
            printf("ERROR in a_g : unknown case : %d (vE=%f) (index_int=%d)\n",case_EGH,vE,index_int);
            printf("case is supposed to be %d\n",case_egh((quadspline_specificPt->int_a_egh[index_int-1][comp]+quadspline_specificPt->int_a_egh[index_int][comp])/2,
                                                        quadspline_paramPt,quadspline_specificPt,comp));
            print_int_eh(quadspline_specificPt,comp);
            if(DEBUG_LEVEL>0){
                printf("ALEX_DEBUG : [END] a_g (comp=%d)\n",comp);
            }
            return quadspline_specificPt->int_a_egh[index_int][comp];
    }
    N=(int)sol[0];
    if(N>0){
        dVmin=fabs(vE-v_e(sol[1],0,quadspline_paramPt,quadspline_specificPt,comp,0));
        aG=sol[1];
        for(k=2;k<=N;k++){
            dV=fabs(vE-v_e(sol[k],0,quadspline_paramPt,quadspline_specificPt,comp,0));
            if(dV<dVmin){
                dVmin=dV;
                aG=sol[k];
            }
        }
    }
    else{
        dV=EPSI+1;
    }
    free(sol);
    if(dV>EPSI){
        mina=quadspline_specificPt->int_a_egh[index_int-1][comp];
        maxa=quadspline_specificPt->int_a_egh[index_int][comp];
        aG=(mina+maxa)/2;
        dV=vE-v_e(aG,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        while(fabs(dV)>EPSI && fabs(maxa-mina)>EPSI){
            if(dV>0){
                mina=aG;
            }
            else{
                maxa=aG;
            }
            aG=(mina+maxa)/2;
            dV=vE-v_e(aG,0,quadspline_paramPt,quadspline_specificPt,comp,0);
        }
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (dicho) (case_EGH=%d)\n",comp,case_EGH);
        }
    }
    else{
        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] a_g (comp=%d) (root) (case_EGH=%d)\n",comp,case_EGH);
        }
    }
    return aG;
}

double d_x(double vD,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] d_x (comp=%d)\n",comp);
    }

    double aB,aG,xC,xE,dX,tD;
    aB=a_b(vD,quadspline_paramPt,quadspline_specificPt,comp);
    durations_and_signs_ac(aB,quadspline_paramPt,quadspline_specificPt,comp);
    aG=a_g(vD,quadspline_paramPt,quadspline_specificPt,comp);
    durations_and_signs_eh(aG,quadspline_paramPt,quadspline_specificPt,comp);
    xC=x_c(aB,0,quadspline_paramPt,quadspline_specificPt,comp,1);
    xE=x_e(aG,0,quadspline_paramPt,quadspline_specificPt,comp,1);
    dX=xE-xC;
    tD=dX/vD;
    if(tD<0 || !isfinite(tD)){
        tD=0;
    }
    quadspline_specificPt->durations[D][comp]=tD;
    quadspline_specificPt->durations[F][comp]=2*quadspline_specificPt->durations[A1][comp]
                                             +quadspline_specificPt->durations[A2][comp]
                                             +quadspline_specificPt->durations[B][comp]
                                             +2*quadspline_specificPt->durations[C1][comp]
                                             +quadspline_specificPt->durations[C2][comp]
                                             +quadspline_specificPt->durations[D][comp]
                                             +2*quadspline_specificPt->durations[E1][comp]
                                             +quadspline_specificPt->durations[E2][comp]
                                             +quadspline_specificPt->durations[G][comp]
                                             +2*quadspline_specificPt->durations[H1][comp]
                                             +quadspline_specificPt->durations[H2][comp];

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] d_x (comp=%d)\n",comp);
    }

    return dX;
}

void v_optim(lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] v_optim (comp=%d)\n",comp);
    }

    double v1,dX1,v2,dX2,v,dX,s,vmax;
    int nb_it=0;
    intervals_ac(quadspline_paramPt,quadspline_specificPt,comp);
    intervals_eh(quadspline_paramPt,quadspline_specificPt,comp);
    vmax=quadspline_paramPt->vmax;
    v1=0;
    dX1=d_x(v1,quadspline_paramPt,quadspline_specificPt,comp);
    s=sign(dX1);
    v2=s*vmax;
    dX2=d_x(v2,quadspline_paramPt,quadspline_specificPt,comp);
    v=v2-(v2-v1)*dX2/(dX2-dX1);
    dX=d_x(v,quadspline_paramPt,quadspline_specificPt,comp);
    while(fabs(dX)>EPSI && nb_it<NB_IT_MAX){
        dX1=dX2;
        dX2=dX;
        v1=v2;
        v2=v;
        v=v2-(v2-v1)*dX2/(dX2-dX1);
        dX=d_x(v,quadspline_paramPt,quadspline_specificPt,comp);
        nb_it++;
    }

    if(fabs(v)>vmax || s*v<0){
        v=s*vmax;
        dX=d_x(v,quadspline_paramPt,quadspline_specificPt,comp);
    }

    if(fabs(dX)>EPSI && s*dX<0){
        v1=0;
        v2=fabs(v);
        v=v2/2;
        dX=d_x(s*v,quadspline_paramPt,quadspline_specificPt,comp);
        while(fabs(dX)>EPSI && fabs(v1-v2)>EPSI){
            if(s*dX<0){
                v2=v;
            }
            else{
                v1=v;
            }
            v=(v1+v2)/2;
            dX=d_x(s*v,quadspline_paramPt,quadspline_specificPt,comp);
        }
        if(s*dX<0){
            v=v1;
            dX=d_x(s*v,quadspline_paramPt,quadspline_specificPt,comp);
        }
        v=s*v;
    }

    quadspline_specificPt->v_opt[comp]=v;

    if(fabs(v)>vmax || s*v<0){
        printf("\nERROR in v_optim : v=%f s=%f vmax=%f\n",v,s,vmax);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] v_optim (comp=%d)\n",comp);
    }

}

void synchronize(double tF_des,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] synchronize (comp=%d)\n",comp);
    }

    double minv,maxv,vD,s,tF,dX,v_o;

    v_o=quadspline_specificPt->v_opt[comp];
    tF=quadspline_specificPt->durations[F][comp];

    if(tF<tF_des){
       s=sign(v_o);
       minv=0;
       maxv=fabs(v_o);
       vD=maxv/2;
       dX=d_x(s*vD,quadspline_paramPt,quadspline_specificPt,comp);
       tF=quadspline_specificPt->durations[F][comp];
       while(fabs(tF-tF_des)>EPSI && fabs(minv-maxv)>EPSI/100){
          if(tF<tF_des){
             maxv=vD;
          }
          else{
             minv=vD;
          }
          vD=(minv+maxv)/2;
          dX=d_x(s*vD,quadspline_paramPt,quadspline_specificPt,comp);
          tF=quadspline_specificPt->durations[F][comp];
       }
       vD=s*vD;
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] synchronize (comp=%d)\n",comp);
    }

}

void spline_at_from_q0(double t,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,double *x,double *v,double *a,double *je){

    int stop,k;
    double t_cur,x_cur,v_cur,a_cur,j_cur,t_rel,t_rel_2,t_rel_3,t_rel_4,x0,v0,a0,smax;
    double T[15];
    double S[15];

    smax=quadspline_paramPt->smax;

    if(comp<3){
        x0=quadspline_specificPt->q_init[quadspline_paramPt->x_dof+comp];
        v0=quadspline_specificPt->q_init[quadspline_paramPt->vx_dof+comp];
        a0=quadspline_specificPt->q_init[quadspline_paramPt->ax_dof+comp];
    }
    else{
        x0=quadspline_specificPt->q_init[quadspline_paramPt->rz_dof];
        v0=0;
        a0=0;
    }

    t_cur=0;
    x_cur=x0;
    v_cur=v0;
    a_cur=a0;
    j_cur=0;

    if(t<0){
        *x=x_cur;
        *v=v_cur;
        *a=a_cur;
        *je=j_cur;
        return;
    }

    stop=0;
    k=-1;

    //T=[tA1 tA2 tA1 tB tC1 tC2 tC1 tD tE1 tE2 tE1 tG tH1 tH2 tH1]
     T[0]=quadspline_specificPt->durations[A1][comp];
     T[1]=quadspline_specificPt->durations[A2][comp];
     T[2]=quadspline_specificPt->durations[A1][comp];
     T[3]=quadspline_specificPt->durations[B][comp];
     T[4]=quadspline_specificPt->durations[C1][comp];
     T[5]=quadspline_specificPt->durations[C2][comp];
     T[6]=quadspline_specificPt->durations[C1][comp];
     T[7]=quadspline_specificPt->durations[D][comp];
     T[8]=quadspline_specificPt->durations[E1][comp];
     T[9]=quadspline_specificPt->durations[E2][comp];
    T[10]=quadspline_specificPt->durations[E1][comp];
    T[11]=quadspline_specificPt->durations[G][comp];
    T[12]=quadspline_specificPt->durations[H1][comp];
    T[13]=quadspline_specificPt->durations[H2][comp];
    T[14]=quadspline_specificPt->durations[H1][comp];

    //S=smax*[sA 0 -sA 0 sC 0 -sC 0 sE 0 -sE 0 sH 0 -sH];
    S[ 0]= smax*quadspline_specificPt->signs[A][comp];
    S[ 1]= 0;
    S[ 2]=-smax*quadspline_specificPt->signs[A][comp];
    S[ 3]= 0;
    S[ 4]= smax*quadspline_specificPt->signs[C][comp];
    S[ 5]= 0;
    S[ 6]=-smax*quadspline_specificPt->signs[C][comp];
    S[ 7]= 0;
    S[ 8]= smax*quadspline_specificPt->signs[E][comp];
    S[ 9]= 0;
    S[10]=-smax*quadspline_specificPt->signs[E][comp];
    S[11]= 0;
    S[12]= smax*quadspline_specificPt->signs[H][comp];
    S[13]= 0;
    S[14]=-smax*quadspline_specificPt->signs[H][comp];

    while(!stop && k<14){
        k=k+1;
        if(t_cur<=t && t<=t_cur+T[k]){
            t_rel=t-t_cur;
            t_rel_2=t_rel*t_rel;
            t_rel_3=t_rel*t_rel_2;
            t_rel_4=t_rel*t_rel_3;
            *je=S[k]*t_rel+j_cur;
            *a=(S[k]/2)*t_rel_2+j_cur*t_rel+a_cur;
            *v=(S[k]/6)*t_rel_3+(j_cur/2)*t_rel_2+a_cur*t_rel+v_cur;
            *x=(S[k]/24)*t_rel_4+(j_cur/6)*t_rel_3+(a_cur/2)*t_rel_2+v_cur*t_rel+x_cur;
            stop=1;
        }
        if(!stop && T[k]>0){
            t_cur=t_cur+T[k];
            t_rel=T[k];
            t_rel_2=t_rel*t_rel;
            t_rel_3=t_rel*t_rel_2;
            t_rel_4=t_rel*t_rel_3;
            x_cur=(S[k]/24)*t_rel_4+(j_cur/6)*t_rel_3+(a_cur/2)*t_rel_2+v_cur*t_rel+x_cur;
            v_cur=(S[k]/6)*t_rel_3+(j_cur/2)*t_rel_2+a_cur*t_rel+v_cur;
            a_cur=(S[k]/2)*t_rel_2+j_cur*t_rel+a_cur;
            j_cur=S[k]*t_rel+j_cur;
        }
    }

    if(!stop){
        *x=x_cur;
        *v=v_cur;
        *a=a_cur;
        *je=j_cur;
    }
}

void spline_at_from_qF(double t,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,double *x,double *v,double *a,double *je){

    int stop,k;
    double t_cur,x_cur,v_cur,a_cur,j_cur,t_rel,t_rel_2,t_rel_3,t_rel_4,tF,xF,vF,aF,smax;
    double T[15];
    double S[15];

    smax=quadspline_paramPt->smax;

    if(comp<3){
        xF=quadspline_specificPt->q_end[quadspline_paramPt->x_dof+comp];
        vF=quadspline_specificPt->q_end[quadspline_paramPt->vx_dof+comp];
        aF=quadspline_specificPt->q_end[quadspline_paramPt->ax_dof+comp];
    }
    else{
        xF=quadspline_specificPt->q_end[quadspline_paramPt->rz_dof];
        vF=0;
        aF=0;
    }

    //T=[tA1 tA2 tA1 tB tC1 tC2 tC1 tD tE1 tE2 tE1 tG tH1 tH2 tH1]
     T[0]=quadspline_specificPt->durations[A1][comp];
     T[1]=quadspline_specificPt->durations[A2][comp];
     T[2]=quadspline_specificPt->durations[A1][comp];
     T[3]=quadspline_specificPt->durations[B][comp];
     T[4]=quadspline_specificPt->durations[C1][comp];
     T[5]=quadspline_specificPt->durations[C2][comp];
     T[6]=quadspline_specificPt->durations[C1][comp];
     T[7]=quadspline_specificPt->durations[D][comp];
     T[8]=quadspline_specificPt->durations[E1][comp];
     T[9]=quadspline_specificPt->durations[E2][comp];
    T[10]=quadspline_specificPt->durations[E1][comp];
    T[11]=quadspline_specificPt->durations[G][comp];
    T[12]=quadspline_specificPt->durations[H1][comp];
    T[13]=quadspline_specificPt->durations[H2][comp];
    T[14]=quadspline_specificPt->durations[H1][comp];

    //S=smax*[sA 0 -sA 0 sC 0 -sC 0 sE 0 -sE 0 sH 0 -sH];
    S[ 0]= smax*quadspline_specificPt->signs[A][comp];
    S[ 1]= 0;
    S[ 2]=-smax*quadspline_specificPt->signs[A][comp];
    S[ 3]= 0;
    S[ 4]= smax*quadspline_specificPt->signs[C][comp];
    S[ 5]= 0;
    S[ 6]=-smax*quadspline_specificPt->signs[C][comp];
    S[ 7]= 0;
    S[ 8]= smax*quadspline_specificPt->signs[E][comp];
    S[ 9]= 0;
    S[10]=-smax*quadspline_specificPt->signs[E][comp];
    S[11]= 0;
    S[12]= smax*quadspline_specificPt->signs[H][comp];
    S[13]= 0;
    S[14]=-smax*quadspline_specificPt->signs[H][comp];

    tF=quadspline_specificPt->durations[F][comp];

    t_cur=tF;
    x_cur=xF;
    v_cur=vF;
    a_cur=aF;
    j_cur=0;

    stop=0;
    k=-1;

    if(t>tF){
        *x=x_cur;
        *v=v_cur;
        *a=a_cur;
        *je=j_cur;
        return;
    }

    while(!stop && k<14){
        k=k+1;
        if(t_cur-T[k]<=t && t<=t_cur){
            t_rel=t-t_cur+T[k];
            t_rel_2=t_rel*t_rel;
            t_rel_3=t_rel*t_rel_2;
            t_rel_4=t_rel*t_rel_3;
            *je=S[k]*t_rel+(j_cur-S[k]*T[k]);
            *a=(S[k]/2)*t_rel_2+(j_cur-S[k]*T[k])*t_rel+(a_cur+(S[k]/2)*t_rel_2-j_cur*T[k]);
            *v=(S[k]/6)*t_rel_3+((j_cur-S[k]*T[k])/2)*t_rel_2+(a_cur+(S[k]/2)*t_rel_2-j_cur*T[k])*t_rel+(v_cur-(S[k]/6)*t_rel_3+(j_cur/2)*t_rel_2-a_cur*T[k]);
            *x=(S[k]/24)*t_rel_4+((j_cur-S[k]*T[k])/6)*t_rel_3+((a_cur+(S[k]/2)*t_rel_2-j_cur*T[k])/2)*t_rel_2+(v_cur-(S[k]/6)*t_rel_3+(j_cur/2)*t_rel_2-a_cur*T[k])*t_rel+(x_cur+(S[k]/24)*t_rel_4-(j_cur/6)*t_rel_3+(a_cur/2)*t_rel_2-v_cur*T[k]);
            stop=1;
        }
        if(!stop && T[k]>0){
            t_cur=t_cur-T[k];
            t_rel=T[k];
            t_rel_2=t_rel*t_rel;
            t_rel_3=t_rel*t_rel_2;
            t_rel_4=t_rel*t_rel_3;
            x_cur=x_cur+(S[k]/24)*t_rel_4-(j_cur/6)*t_rel_3+(a_cur/2)*t_rel_2-v_cur*T[k];
            v_cur=v_cur-(S[k]/6)*t_rel_3+(j_cur/2)*t_rel_2-a_cur*T[k];
            a_cur=a_cur+(S[k]/2)*t_rel_2-j_cur*T[k];
            j_cur=j_cur-S[k]*T[k];
        }
    }

    if(!stop){
        *x=x_cur;
        *v=v_cur;
        *a=a_cur;
        *je=j_cur;
    }
}

void spline_at(double t,lm_quadspline_str *quadspline_paramPt,p3d_quadspline_data *quadspline_specificPt,int comp,double *x,double *v,double *a,double *je){
    /*
    int k;
    double tCD=0;
    for(k=0;k<5;k++){
        tCD=tCD+quadspline_specificPt->durations[k][comp];
    }
    tCD=tCD+quadspline_specificPt->durations[A1][comp]+quadspline_specificPt->durations[C1][comp];
    if(t<=tCD){
    */
        spline_at_from_q0(t,quadspline_paramPt,quadspline_specificPt,comp,x,v,a,je);
    /*
    }
    else{
        spline_at_from_qF(t,quadspline_paramPt,quadspline_specificPt,comp,x,v,a,je);
    }
    */
}

/* allocation of a data structure specific to the quadspline local method */
p3d_quadspline_data * p3d_alloc_spec_quadspline_localpath(p3d_rob *robotPt,
                                                          configPt q_i,
                                                          configPt q_f){

    p3d_quadspline_data *quadspline_specificPt;
    int comp,slowest_comp=-1,ERROR,k;
    double tF_des;
    double tF[4],v_opt[4];
    double x0,y0,z0,xF,yF,zF,dist;

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_alloc_spec_quadspline_local_path\n");
    }

    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_alloc_spec_quadspline_localpath : \n");
        printf("quadspline_specificPt is NULL\n");
        return NULL;
    }

    if ((quadspline_specificPt = MY_ALLOC(p3d_quadspline_data,1)) == NULL)
        return NULL;

    quadspline_specificPt->q_init = q_i;
    quadspline_specificPt->q_end = q_f;
    quadspline_specificPt->calculated=TRUE;

    if(fabs(q_i[quadspline_paramPt->vx_dof])<EPS10 &&
       fabs(q_i[quadspline_paramPt->vy_dof])<EPS10 &&
       fabs(q_i[quadspline_paramPt->vz_dof])<EPS10 &&
       fabs(q_i[quadspline_paramPt->ax_dof])<EPS10 &&
       fabs(q_i[quadspline_paramPt->ay_dof])<EPS10 &&
       fabs(q_i[quadspline_paramPt->az_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->vx_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->vy_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->vz_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->ax_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->ay_dof])<EPS10 &&
       fabs(q_f[quadspline_paramPt->az_dof])<EPS10){
        if(fabs(q_i[quadspline_paramPt->x_dof]-q_f[quadspline_paramPt->x_dof])<EPS10 &&
           fabs(q_i[quadspline_paramPt->y_dof]-q_f[quadspline_paramPt->y_dof])<EPS10 &&
           fabs(q_i[quadspline_paramPt->z_dof]-q_f[quadspline_paramPt->z_dof])<EPS10 &&
           fabs(q_i[quadspline_paramPt->rz_dof]-q_f[quadspline_paramPt->rz_dof])<EPS10){
            p3d_matrix4 absPos;
            configPt q=p3d_alloc_config(robotPt);
            q[quadspline_paramPt->x_dof]=q_i[quadspline_paramPt->x_dof];
            q[quadspline_paramPt->y_dof]=q_i[quadspline_paramPt->y_dof];
            q[quadspline_paramPt->z_dof]=q_i[quadspline_paramPt->z_dof];
            q[quadspline_paramPt->rx_dof]=0;
            q[quadspline_paramPt->ry_dof]=0;
            q[quadspline_paramPt->rz_dof]=q_i[quadspline_paramPt->rz_dof];
            q[quadspline_paramPt->vx_dof]=0;
            q[quadspline_paramPt->vy_dof]=0;
            q[quadspline_paramPt->vz_dof]=0;
            q[quadspline_paramPt->ax_dof]=0;
            q[quadspline_paramPt->ay_dof]=0;
            q[quadspline_paramPt->az_dof]=0;
            dist=0;
            p3d_point EF_i,EF_f;
            p3d_set_and_update_this_robot_conf(robotPt,q_i);
            p3d_mat4Copy(robotPt->joints[EF_JNT]->abs_pos,absPos);
            EF_i.x=absPos[0][3];
            EF_i.y=absPos[1][3];
            EF_i.z=absPos[2][3];
            for(double param=0.01;param<=1;param+=0.01){
                for(k=quadspline_paramPt->az_dof+1;k<robotPt->nb_dof;k++){
                    q[k]=param*(q_f[k]-q_i[k])+q_i[k];
                }
                p3d_set_and_update_this_robot_conf(robotPt,q);
                p3d_mat4Copy(robotPt->joints[EF_JNT]->abs_pos,absPos);
                EF_f.x=absPos[0][3];
                EF_f.y=absPos[1][3];
                EF_f.z=absPos[2][3];
                dist+=sqrt(SQR(EF_f.x-EF_i.x)+SQR(EF_f.y-EF_i.y)+SQR(EF_f.z-EF_i.z));
                EF_i.x=EF_f.x;
                EF_i.y=EF_f.y;
                EF_i.z=EF_f.z;
            }
            double maxt=0;
            for(k=quadspline_paramPt->az_dof+1;k<robotPt->nb_dof;k++){
                double t=fabs(q_f[robotPt->nb_dof-1]-q_i[robotPt->nb_dof-1])/JNTMAXANGVEL;
                if(t>maxt) maxt=t;
            }
            quadspline_specificPt->tf=std::max(dist/EFLINVEL,maxt);
        }
        else{
            x0=q_i[quadspline_paramPt->x_dof];
            y0=q_i[quadspline_paramPt->y_dof];
            z0=q_i[quadspline_paramPt->z_dof];
            xF=q_f[quadspline_paramPt->x_dof];
            yF=q_f[quadspline_paramPt->y_dof];
            zF=q_f[quadspline_paramPt->z_dof];
            dist=sqrt(SQR(xF-x0)+SQR(yF-y0)+SQR(zF-z0));
            q_i[quadspline_paramPt->x_dof]=0;
            q_f[quadspline_paramPt->x_dof]=dist;
            v_optim(quadspline_paramPt,quadspline_specificPt,0);
            v_opt[0]=quadspline_specificPt->v_opt[0];
            tF[0]=quadspline_specificPt->durations[F][0];
            v_optim(quadspline_paramPt,quadspline_specificPt,3);
            v_opt[3]=quadspline_specificPt->v_opt[3];
            tF[3]=quadspline_specificPt->durations[F][3];
            if(tF[3]>tF[0]){
                quadspline_specificPt->v_opt[0]=v_opt[0];
                quadspline_specificPt->durations[F][0]=tF[0];
                synchronize(tF[3],quadspline_paramPt,quadspline_specificPt,0);
                quadspline_specificPt->tf=tF[3];
            }
            else{
                quadspline_specificPt->v_opt[3]=v_opt[3];
                quadspline_specificPt->durations[F][3]=tF[3];
                synchronize(tF[0],quadspline_paramPt,quadspline_specificPt,3);
                quadspline_specificPt->tf=tF[0];
            }
            q_i[quadspline_paramPt->x_dof]=x0;
            q_f[quadspline_paramPt->x_dof]=xF;
            for(comp=1;comp<3;comp++){
                for(k=0;k<12;k++){
                    quadspline_specificPt->durations[k][comp]=quadspline_specificPt->durations[k][0];
                }
                for(k=0;k<4;k++){
                    quadspline_specificPt->signs[k][comp]=quadspline_specificPt->signs[k][0];
                }
            }
        }
    }
    else{
        tF_des=0;
        for(comp=0;comp<4;comp++){
            v_optim(quadspline_paramPt,quadspline_specificPt,comp);
            v_opt[comp]=quadspline_specificPt->v_opt[comp];
            tF[comp]=quadspline_specificPt->durations[F][comp];
            if(tF[comp]>tF_des){
                tF_des=tF[comp];
                slowest_comp=comp;
            }
        }

        //printf("ALEX_DEBUG : tF_des=%f\n",tF_des);

        ERROR=0;
        for(comp=0;comp<4;comp++){
            ERROR=(ERROR || (fabs(quadspline_specificPt->v_opt[comp])>quadspline_paramPt->vmax));
        }

        for(comp=0;comp<4;comp++){
            quadspline_specificPt->v_opt[comp]=v_opt[comp];
            quadspline_specificPt->durations[F][comp]=tF[comp];
            synchronize(tF_des,quadspline_paramPt,quadspline_specificPt,comp);
            //printf("ALEX_DEBUG : tF[%d]=%f\n",comp,quadspline_specificPt->durations[F][comp]);
        }
        quadspline_specificPt->tf=tF_des;
    }

    if(DEBUG_LEVEL>0){
        //printf("ALEX_DEBUG : [END] p3d_alloc_spec_quadspline_local_path\n");
    }

    return quadspline_specificPt;
}

/* allocation of local path of type quadspline */
p3d_localpath * p3d_alloc_quadspline_localpath(p3d_rob *robotPt,
                    configPt q_i, configPt q_f,
                    int lp_id,int is_valid){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_alloc_quadspline_localpath\n");
    }

    p3d_localpath * localpathPt = NULL;

    if ((localpathPt = MY_ALLOC(p3d_localpath, 1)) == NULL)
        return NULL;

    /* allocation of the specific part */
    localpathPt->specific.quadspline_data = p3d_alloc_spec_quadspline_localpath(
                                                               robotPt,q_i,q_f);

    if (localpathPt->specific.quadspline_data == NULL){
        /* allocation failed free everything and return NULL*/
        MY_FREE(localpathPt, p3d_localpath, 1);
        return NULL;
    }

    /* Modification of the attitude of q_i and q_f */

    double Tx,Ty,Tz,m,f,g,ry,rx,vx,vy,vz,ax,ay,az;

    lm_quadspline_str *quadspline_paramPt;
    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_alloc_quadspline_localpath : ");
        printf("quadspline_paramPt is NULL\n");
        return NULL;
    }

    m=quadspline_paramPt->mass;
    f=quadspline_paramPt->friction;
    g=quadspline_paramPt->gravity;

    vx=q_i[quadspline_paramPt->vx_dof];
    vy=q_i[quadspline_paramPt->vy_dof];
    vz=q_i[quadspline_paramPt->vz_dof];

    ax=q_i[quadspline_paramPt->ax_dof];
    ay=q_i[quadspline_paramPt->ay_dof];
    az=q_i[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;

    if(!attitude(Tx,Ty,Tz,&rx,&ry)) printf("ERROR in attitude\n");

    q_i[quadspline_paramPt->rx_dof]=rx;
    q_i[quadspline_paramPt->ry_dof]=ry;

    vx=q_f[quadspline_paramPt->vx_dof];
    vy=q_f[quadspline_paramPt->vy_dof];
    vz=q_f[quadspline_paramPt->vz_dof];

    ax=q_f[quadspline_paramPt->ax_dof];
    ay=q_f[quadspline_paramPt->ay_dof];
    az=q_f[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;

    if(!attitude(Tx,Ty,Tz,&rx,&ry)) printf("ERROR in attitude\n");

    q_f[quadspline_paramPt->rx_dof]=rx;
    q_f[quadspline_paramPt->ry_dof]=ry;

    if (localpathPt->specific.quadspline_data == NULL){
        /* allocation failed free everything and return NULL*/
        MY_FREE(localpathPt, p3d_localpath, 1);
        return NULL;
    }

    /* Initialization of the generic part */

    /* fields */
    localpathPt->type_lp = QUADSPLINE;
    localpathPt->valid = is_valid;
    localpathPt->lp_id = lp_id;
    localpathPt->prev_lp = NULL;
    localpathPt->next_lp = NULL;
    localpathPt->length_lp=localpathPt->specific.quadspline_data->tf;
    localpathPt->range_param=localpathPt->specific.quadspline_data->tf;

    /* methods associated to the local path */

    /* compute the length of the local path */
    localpathPt->length =
        (double (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_dist);

    /* computes the configuration at given parameter along the path */
    localpathPt->config_at_param =
        (configPt (*)(p3d_rob*, p3d_localpath*,double))
                                               (p3d_quadspline_config_at_param);
    localpathPt->config_at_distance =
        (configPt (*)(p3d_rob*, p3d_localpath*,double))
                                               (p3d_quadspline_config_at_param);
    localpathPt->extract_sub_localpath =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,double, double))
                                                       (p3d_extract_quadspline);

    /* extract from a local path a sub local path starting from parameter u1
       and ending at parameter u2 */
    localpathPt->extract_by_param =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*,double, double))
                                                       (p3d_extract_quadspline);

    /* destroy the localpath */
    localpathPt->destroy =
        (void (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_destroy);

    /*copy the local path */
    localpathPt->copy =
        (p3d_localpath* (*)(p3d_rob*,p3d_localpath*))
                                                (p3d_copy_quadspline_localpath);

    /* compute the cost of a local path */
    localpathPt->cost =
        (double (*)(p3d_rob*, p3d_localpath*))(p3d_quadspline_cost);

    /* function that simplifies the sequence of two local paths :
       valid only for RS curves */
    localpathPt->simplify =
        (p3d_localpath* (*)(p3d_rob*, p3d_localpath*, int*))
                                                      (p3d_simplify_quadspline);

    /* write the local path in a file */
    localpathPt->write =
        (int (*)(FILE *, p3d_rob*, p3d_localpath*))(p3d_write_quadspline);

    /* from a configuration on a local path, this function computes an
     interval of parameter on the local path on which all the points
     of the robot move by less than the distance given as input.
     The interval is centered on the configuration given as input. The
     function returns the half length of the interval */
    localpathPt->stay_within_dist =
        (double (*)(p3d_rob*, p3d_localpath*,double, whichway, double*))
                                              (p3d_quadspline_stay_within_dist);

    localpathPt->ikSol = NULL;

    //save the active constraints
    localpathPt->nbActiveCntrts = 0;
    localpathPt->activeCntrts = NULL;

    #ifdef LIGHT_PLANNER
        localpathPt->isCarryingObject = FALSE;
        for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
            localpathPt->carriedObject[i] = NULL;
        }
    #endif

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_alloc_quadspline_localpath\n");
    }

    return localpathPt;
}

/* distance for the quadspline local method
 * The output is actually the time of the motion between qi and qf
 * along the local path
 * */
double p3d_quadspline_dist(p3d_rob *robotPt, p3d_localpath *localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_dist\n");
    }

    /* cast the pointer to union p3d_lm_specific to a pointer
     to p3d_quadspline_data */
    p3d_quadspline_data *specificPt = localpathPt->specific.quadspline_data;

    /* test whether the type of local path is the expected one */
    if (localpathPt->type_lp != QUADSPLINE){
        printf("p3d_quadspline_dist: Quadspline local path expected\n");
    }

    if(specificPt==NULL && DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : specificPt is NULL in p3d_quadspline_dist\n");
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_dist\n");
    }

    return specificPt->tf;
}

/*
 *  Compute the configuration situated at given parameter on the local path.
 *
 *  Input:  the robot, the distance.
 *
 *  Output: the configuration
 */
configPt p3d_quadspline_config_at_param(p3d_rob *robotPt,
                                        p3d_localpath *localpathPt,
                                        double param){

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_config_at_param\n");
    }

    p3d_quadspline_data *quadspline_specificPt;
    lm_quadspline_str *quadspline_paramPt;
    configPt q,q_init,q_end;
    double Tx,Ty,Tz,m,f,g,rx,ry,vx,vy,vz,ax,ay,az,x,v,a,je;
    double x0,y0,z0,xF,yF,zF,dist,dx,dy,dz;
    int comp,k;

    if (localpathPt == NULL){
        printf("ERROR in p3d_alloc_quadspline_localpath : ");
        printf("localpathPt is NULL\n");
        return NULL;
    }

    if (localpathPt->type_lp != QUADSPLINE){
        printf("ERROR in p3d_quadspline_config_at_param : ");
        printf("local path must be of type QUADSPLINE\n");
        return NULL;
    }

    quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in p3d_alloc_quadspline_localpath : ");
        printf("quadspline_paramPt is NULL\n");
        return NULL;
    }

    quadspline_specificPt = localpathPt->specific.quadspline_data;
    if (quadspline_specificPt == NULL){
        printf("ERROR in p3d_alloc_quadspline_localpath : ");
        printf("quadspline_specificPt is NULL\n");
        return NULL;
    }

    q_init=quadspline_specificPt->q_init;
    q_end=quadspline_specificPt->q_end;

    if (param<0 || fabs(param)<EPS10 || fabs(quadspline_specificPt->tf)<EPS10){
        //printf("COPY q_init\n");
        return p3d_copy_config(robotPt,q_init);
    }

    if (param>quadspline_specificPt->tf || fabs(param-quadspline_specificPt->tf)<EPS10){
        //printf("COPY q_end\n");
        return p3d_copy_config(robotPt,q_end);
    }

    //printf("EVAL\n");

    q=p3d_alloc_config(robotPt);

    if(fabs(q_init[quadspline_paramPt->vx_dof])<EPS10 &&
       fabs(q_init[quadspline_paramPt->vy_dof])<EPS10 &&
       fabs(q_init[quadspline_paramPt->vz_dof])<EPS10 &&
       fabs(q_init[quadspline_paramPt->ax_dof])<EPS10 &&
       fabs(q_init[quadspline_paramPt->ay_dof])<EPS10 &&
       fabs(q_init[quadspline_paramPt->az_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->ax_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->ay_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->az_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->ax_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->ay_dof])<EPS10 &&
       fabs(q_end[quadspline_paramPt->az_dof])<EPS10){
        if(fabs(q_init[quadspline_paramPt->x_dof]-q_end[quadspline_paramPt->x_dof])<EPS10 &&
           fabs(q_init[quadspline_paramPt->y_dof]-q_end[quadspline_paramPt->y_dof])<EPS10 &&
           fabs(q_init[quadspline_paramPt->z_dof]-q_end[quadspline_paramPt->z_dof])<EPS10 &&
           fabs(q_init[quadspline_paramPt->rz_dof]-q_end[quadspline_paramPt->rz_dof])<EPS10){
            q[quadspline_paramPt->x_dof]=q_init[quadspline_paramPt->x_dof];
            q[quadspline_paramPt->y_dof]=q_init[quadspline_paramPt->y_dof];
            q[quadspline_paramPt->z_dof]=q_init[quadspline_paramPt->z_dof];
            q[quadspline_paramPt->rx_dof]=0;
            q[quadspline_paramPt->ry_dof]=0;
            q[quadspline_paramPt->rz_dof]=q_init[quadspline_paramPt->rz_dof];
            q[quadspline_paramPt->vx_dof]=0;
            q[quadspline_paramPt->vy_dof]=0;
            q[quadspline_paramPt->vz_dof]=0;
            q[quadspline_paramPt->ax_dof]=0;
            q[quadspline_paramPt->ay_dof]=0;
            q[quadspline_paramPt->az_dof]=0;
            for(k=0;k<quadspline_paramPt->x_dof;k++){
                q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
            }
            for(k=quadspline_paramPt->az_dof+1;k<robotPt->nb_dof;k++){
                q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
            }
            return q;
        }
        else{
            x0=q_init[quadspline_paramPt->x_dof];
            y0=q_init[quadspline_paramPt->y_dof];
            z0=q_init[quadspline_paramPt->z_dof];
            xF=q_end[quadspline_paramPt->x_dof];
            yF=q_end[quadspline_paramPt->y_dof];
            zF=q_end[quadspline_paramPt->z_dof];
            dist=sqrt(SQR(xF-x0)+SQR(yF-y0)+SQR(zF-z0));
            q_init[quadspline_paramPt->x_dof]=0;
            q_end[quadspline_paramPt->x_dof]=dist;
            spline_at(param,quadspline_paramPt,quadspline_specificPt,0,&x,&v,&a,&je);
            q_init[quadspline_paramPt->x_dof]=x0;
            q_end[quadspline_paramPt->x_dof]=xF;
            dx=(xF-x0)/dist;
            q[quadspline_paramPt->x_dof]=x0+x*dx;
            q[quadspline_paramPt->vx_dof]=v*dx;
            q[quadspline_paramPt->ax_dof]=a*dx;
            dy=(yF-y0)/dist;
            q[quadspline_paramPt->y_dof]=y0+x*dy;
            q[quadspline_paramPt->vy_dof]=v*dy;
            q[quadspline_paramPt->ay_dof]=a*dy;
            dz=(zF-z0)/dist;
            q[quadspline_paramPt->z_dof]=z0+x*dz;
            q[quadspline_paramPt->vz_dof]=v*dz;
            q[quadspline_paramPt->az_dof]=a*dz;
            spline_at(param,quadspline_paramPt,quadspline_specificPt,3,&x,&v,&a,&je);
            q[quadspline_paramPt->rz_dof]=x;
        }
    }
    else{
        for(comp=0;comp<4;comp++){
            spline_at(param,quadspline_paramPt,quadspline_specificPt,comp,&x,&v,&a,&je);
            //printf("ALEX_DEBUG : spline_at : x=%f\n",x);
            if(comp<3){
                q[quadspline_paramPt->x_dof+comp]=x;
                q[quadspline_paramPt->vx_dof+comp]=v;
                q[quadspline_paramPt->ax_dof+comp]=a;
            }
            else{
                q[quadspline_paramPt->rz_dof]=x;
            }
        }
    }

    m=quadspline_paramPt->mass;
    g=quadspline_paramPt->gravity;
    f=quadspline_paramPt->friction;

    m=quadspline_paramPt->mass;
    f=quadspline_paramPt->friction;
    g=quadspline_paramPt->gravity;

    vx=q[quadspline_paramPt->vx_dof];
    vy=q[quadspline_paramPt->vy_dof];
    vz=q[quadspline_paramPt->vz_dof];

    ax=q[quadspline_paramPt->ax_dof];
    ay=q[quadspline_paramPt->ay_dof];
    az=q[quadspline_paramPt->az_dof];

    Tx=m*ax+f*vx;
    Ty=m*ay+f*vy;
    Tz=m*az+f*vz+m*g;

    if(!attitude(Tx,Ty,Tz,&rx,&ry)) printf("ERROR in attitude\n");

    q[quadspline_paramPt->rx_dof]=rx;
    q[quadspline_paramPt->ry_dof]=ry;

    for(k=0;k<quadspline_paramPt->x_dof;k++){
        q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
    }
    for(k=quadspline_paramPt->az_dof+1;k<robotPt->nb_dof;k++){
        q[k]=param*(q_end[k]-q_init[k])/quadspline_specificPt->tf+q_init[k];
    }

    if(DEBUG_LEVEL!=0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_config_at_param\n");
    }

    return q;
}

/*
 *  Extract from an quadspline local path the sub local path starting
 *  at parameter t1 and ending at parameter t2.
 *  The length of the extracted local path is computed
 *
 *  If t2 > length local path, return end of local path
 */
p3d_localpath *p3d_extract_quadspline(p3d_rob *robotPt,
                                      p3d_localpath *localpathPt,
                                      double t1, double t2){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_extract_quadspline\n");
    }

    configPt q1, q2;
    p3d_localpath *sub_localpathPt;

    q1 = p3d_quadspline_config_at_param(robotPt, localpathPt, t1);
    if (robotPt->cntrt_manager->cntrts != NULL) {
        p3d_set_and_update_this_robot_conf_multisol(robotPt, q1, NULL, 0,
                                                    localpathPt->ikSol);
        p3d_get_robot_config_into(robotPt, &q1);
    }

    q2 = p3d_quadspline_config_at_param(robotPt, localpathPt, t2);
    if (robotPt->cntrt_manager->cntrts != NULL) {
        p3d_set_and_update_this_robot_conf_multisol(robotPt, q2, NULL, 0,
                                                    localpathPt->ikSol);
        p3d_get_robot_config_into(robotPt, &q2);
    }

    sub_localpathPt = p3d_alloc_quadspline_localpath(robotPt, q1, q2, FORWARD,
                                                     localpathPt->valid);

    p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol,
                   &(sub_localpathPt->ikSol));
    sub_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
    sub_localpathPt->activeCntrts = MY_ALLOC(int,
                                             sub_localpathPt->nbActiveCntrts);
    for(int i = 0; i < sub_localpathPt->nbActiveCntrts; i++){
        sub_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_extract_quadspline\n");
    }

    return sub_localpathPt;
}

/*
 *  lm_get_quadspline_lm_param --
 *
 *  find the first occurrence of quadspline local method parameters with same robot.
 */

lm_quadspline_str *lm_get_quadspline_lm_param(p3d_rob* robotPt){

    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [START] lm_get_quadspline_lm_param\n");
    }

    if(!robotPt){
        printf("error: lm_get_quadspline_lm_param: robot parameter is NULL\n");
        return NULL;
    }

    lm_list_param_str *list_paramPt = robotPt->local_method_params;
    lm_quadspline_str *resultPt = NULL;
    while (list_paramPt!=NULL) {
        if (list_paramPt->lpl_type != QUADSPLINE) {
            list_paramPt = list_paramPt->next;
        }
        else{
            resultPt = (plm_quadspline_str)(list_paramPt->lm_param);
            if(resultPt->robotPt==robotPt){
                if(DEBUG_LEVEL>1){
                    printf("ALEX_DEBUG : [END] lm_get_quadspline_lm_param\n");
                }
                return resultPt;
            }
            else{
                resultPt=NULL;
                list_paramPt = list_paramPt->next;
            }
        }
    }
    if(DEBUG_LEVEL>1){
        printf("ALEX_DEBUG : [END (NULL)] lm_get_quadspline_lm_param\n");
    }
    if(!resultPt)
        printf("error: lm_get_quadspline_lm_param: lm_quadspline_str has not been initialized for robot %s\n",robotPt->name);

    return resultPt;
}

/*
 *  lm_create_quadspline
 *
 *  create a Quadspline local method structure
 *
 *  Input: dtab[0,1,2,3,5,6,7] = vmax,amax,jmax,smax,mass,friction,gravity
 *         itab[0,1] = num_joint
 */

lm_quadspline_str *lm_create_quadspline(p3d_rob *robotPt, double *dtab,
                                        int *itab){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] lm_create_quadspline\n");
    }

    lm_quadspline_str* quadspline_paramPt = NULL;

    if ((quadspline_paramPt = MY_ALLOC(lm_quadspline_str, 1)) == NULL) {
        printf("ERROR in lm_create_quadspline : allocation failed\n");
        return (NULL);
    }

    if(dtab[0]<=0){
        printf("ERROR in lm_create_quadspline : vmax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->vmax = dtab[0];

    if(dtab[1]<=0){
        printf("ERROR in lm_create_quadspline : amax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->amax = dtab[1];

    if(dtab[2]<=0){
        printf("ERROR in lm_create_quadspline : jmax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->jmax = dtab[2];

    if(dtab[3]<=0){
        printf("ERROR in lm_create_quadspline : smax is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->smax = dtab[3];

    if(dtab[4]<=0){
        printf("ERROR in lm_create_quadspline : mass is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->mass = dtab[4];

    if(dtab[5]<0){
        printf("ERROR in lm_create_quadspline : ");
        printf("friction is negative\n");
        return NULL;
    }
    quadspline_paramPt->friction = dtab[5];

    if(dtab[6]<=0){
        printf("ERROR in lm_create_quadspline : gravity is negative or null\n");
        return NULL;
    }
    quadspline_paramPt->gravity = dtab[6];

    if(itab[0]<0 || itab[0]>robotPt->njoints){
        printf("ERROR in lm_create_quadspline : num_joint is out of range\n");
        return NULL;
    }
    quadspline_paramPt->num_joint = itab[0];

    int k;

    p3d_jnt* jnt=robotPt->joints[quadspline_paramPt->num_joint];
    if(jnt->type!=P3D_QUADROTOR){
        printf("ERROR in lm_create_quadspline : ");
        printf("joint of type P3D_QUADROTOR expected\n");
        return NULL;
    }

    for(k=6;k<9;k++){
        jnt->dof_data[k].vmin=-quadspline_paramPt->vmax;
        jnt->dof_data[k].vmax= quadspline_paramPt->vmax;
    }
    for(k=9;k<12;k++){
        jnt->dof_data[k].vmin=-quadspline_paramPt->amax;
        jnt->dof_data[k].vmax= quadspline_paramPt->amax;
    }

    quadspline_paramPt->x_dof=jnt->index_dof;
    quadspline_paramPt->y_dof=jnt->index_dof+1;
    quadspline_paramPt->z_dof=jnt->index_dof+2;
    quadspline_paramPt->rx_dof=jnt->index_dof+3;
    quadspline_paramPt->ry_dof=jnt->index_dof+4;
    quadspline_paramPt->rz_dof=jnt->index_dof+5;
    quadspline_paramPt->vx_dof=jnt->index_dof+6;
    quadspline_paramPt->vy_dof=jnt->index_dof+7;
    quadspline_paramPt->vz_dof=jnt->index_dof+8;
    quadspline_paramPt->ax_dof=jnt->index_dof+9;
    quadspline_paramPt->ay_dof=jnt->index_dof+10;
    quadspline_paramPt->az_dof=jnt->index_dof+11;

    quadspline_paramPt->robotPt=robotPt;

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] lm_create_quadspline\n");
    }

    return(quadspline_paramPt);
}

/*
 *  p3d_create_quadspline_local_method --
 *
 *  create data-structure to store parameters of quadspline local method
 */
int p3d_create_quadspline_local_method(char *name, double *dtab, int *itab){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_create_quadspline_local_method\n");
    }

    p3d_rob *robotPt=NULL;
    if(name) robotPt=p3d_get_robot_by_name(name);
    if(!robotPt){
        printf("error: p3d_create_quadspline_local_method: bad name for robotPt\n");
        return FALSE;
    }

    robotPt->lpl_type = QUADSPLINE;
    plm_quadspline_str quadspline_paramPt;
    quadspline_paramPt = lm_create_quadspline(robotPt, dtab, itab);

    if (quadspline_paramPt != NULL) {
        robotPt->local_method_params =
            lm_append_to_list(robotPt->local_method_params,
                              (void*)quadspline_paramPt, QUADSPLINE);
    }

    printf("lm_quadspline_str initialized for robot %s\n",name);

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_create_quadspline_local_method\n");
    }

    return TRUE;
}

/*
 * destroys a structure of type p3d_quadspline_data
 */
void p3d_destroy_quadspline_data(p3d_rob* robotPt,
                                 p3d_quadspline_data* quadspline_dataPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_destroy_quadspline_data\n");
    }

    if (quadspline_dataPt != NULL){
        if (quadspline_dataPt->q_init != NULL){
            p3d_destroy_config(robotPt, quadspline_dataPt->q_init);
        }
        if (quadspline_dataPt->q_end != NULL){
            p3d_destroy_config(robotPt, quadspline_dataPt->q_end);
        }
        MY_FREE(quadspline_dataPt, p3d_quadspline_data, 1);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_destroy_quadspline_data\n");
    }

}

/*
 * destroy a Quadspline local path
 */
void p3d_quadspline_destroy(p3d_rob* robotPt, p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_destroy\n");
    }

    if (localpathPt != NULL){
        /* test whether the type of local path is the expected one */
        if (localpathPt->type_lp != QUADSPLINE){
            printf("ERROR in p3d_quadspline_destroy : ");
            printf("Quadspline local path expected\n");
        }
        /* destroy the specific part */
        if (localpathPt->specific.quadspline_data != NULL){
            p3d_destroy_quadspline_data(robotPt,
                                        localpathPt->specific.quadspline_data);
        }
        localpathPt->next_lp = NULL;
        localpathPt->prev_lp = NULL;
        if (localpathPt->ikSol){
            p3d_destroy_specific_iksol(robotPt->cntrt_manager,
                                       localpathPt->ikSol);
            localpathPt->ikSol = NULL;
        }
        MY_FREE(localpathPt->activeCntrts, int, localpathPt->nbActiveCntrts);
        MY_FREE(localpathPt, p3d_localpath, 1);
    }

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_destroy\n");
    }

}

/*
 *  Copy one local path.
 *
 *  Input:  the robot, the local path.
 *
 *  Output: the copied local path
 */

p3d_localpath *p3d_copy_quadspline_localpath(p3d_rob* robotPt,
                                             p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_copy_quadspline_localpath\n");
    }

    configPt q_i, q_f;
    q_i=p3d_copy_config(robotPt, localpathPt->specific.quadspline_data->q_init);
    q_f=p3d_copy_config(robotPt, localpathPt->specific.quadspline_data->q_end);

    p3d_localpath* copy_localpathPt=p3d_alloc_quadspline_localpath(robotPt,q_i,
                                                                   q_f,0,TRUE);

    /* update length and range of parameter */
    copy_localpathPt->length_lp = localpathPt->specific.quadspline_data->tf;
    copy_localpathPt->range_param = localpathPt->specific.quadspline_data->tf;
    p3d_copy_iksol(robotPt->cntrt_manager, localpathPt->ikSol,
                   &(copy_localpathPt->ikSol));
    copy_localpathPt->nbActiveCntrts = localpathPt->nbActiveCntrts;
    copy_localpathPt->activeCntrts = MY_ALLOC(int,
                                              copy_localpathPt->nbActiveCntrts);
    for(int i = 0; i < copy_localpathPt->nbActiveCntrts; i++){
        copy_localpathPt->activeCntrts[i] = localpathPt->activeCntrts[i];
    }

    #ifdef LIGHT_PLANNER
        copy_localpathPt->isCarryingObject = localpathPt->isCarryingObject;
        for (int i = 0; i < MAX_CARRIED_OBJECTS; i++) {
            copy_localpathPt->carriedObject[i] = localpathPt->carriedObject[i];
        }
    #endif

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_copy_quadspline_localpath\n");
    }

    return copy_localpathPt;
}

/*
 *  Cost of a local path
 *
 *  Input:  the local path
 *
 *  Output: the cost
 */
double p3d_quadspline_cost(p3d_rob *robotPt, p3d_localpath *localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_cost\n");
        printf("ALEX_DEBUG : [END] p3d_quadspline_cost\n");
    }

    //return localpathPt->specific.quadspline_data->tf;
    if(fabs(localpathPt->specific.quadspline_data->tf-localpathPt->range_param)>1e-3) printf("!!!!!!!!!!!!!! ERROR !!!!!!!!!!!\n");
    return localpathPt->range_param;
}

/*
 *  does nothing
 */
p3d_localpath *p3d_simplify_quadspline(p3d_rob *robotPt,
                               p3d_localpath *localpathPt, int *need_colcheck){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_simplify_quadspline\n");
        printf("ALEX_DEBUG : [END] p3d_simplify_quadspline\n");
    }

    return localpathPt;
}

/*
 *  p3d_write_quadspline --
 *
 *  write a localpath of type quadspline in a file
 *
 *  ARGS IN  : a file descriptor,
 *             a robot,
 *             a localpath
 *
 *  ARGS OUT : TRUE if success,
 *             FALSE if input local path is not an quadspline one.
 */
int p3d_write_quadspline(FILE *file, p3d_rob* robotPt,
                         p3d_localpath* localpathPt){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_write_quadspline\n");
    }

    p3d_quadspline_data *quadspline_dataPt = NULL;

    if (localpathPt->type_lp != QUADSPLINE) {
        return FALSE;
    }

    fprintf(file, "\n\np3d_add_localpath quadspline\n");

    quadspline_dataPt =
                    (pp3d_quadspline_data)localpathPt->specific.quadspline_data;
    /* write each RS segment */
    fprintf(file, "conf_init");
    fprint_config_one_line(file, robotPt, quadspline_dataPt->q_init);
    fprintf(file, "\nconf_end ");
    fprint_config_one_line(file, robotPt, quadspline_dataPt->q_end);
    fprintf(file, "\ntf %f",quadspline_dataPt->tf);
    fprintf(file, "\np3d_end_local_path\n");

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_write_quadspline\n");
    }

    return TRUE;
}

/*  p3d_quadspline_stay_within_dist
 *
 *  Input:  the robot,
 *          the local path,
 *          the parameter along the curve,
 *          the direction of motion
 *          the maximal distance moved by all the points of the
 *          robot
 *
 *  Output: length of the interval of parameter the robot can
 *          stay on without any body moving by more than the input distance
 *
 *  Description:
 *          From a configuration on a local path, this function
 *          computes an interval of parameter on the local path on
 *          which all the points of the robot move by less than the
 *          distance given as input.  The interval is centered on the
 *          configuration given as input. The function returns the
 *          half length of the interval
 */
double p3d_quadspline_stay_within_dist(p3d_rob* robotPt,
                p3d_localpath* localpathPt,
                double parameter, whichway dir,
                double *distances){




    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_stay_within_dist\n");
    }

  p3d_quadspline_data *quadspline_localpathPt=NULL;
  int i, j, njnt = robotPt->njoints;
  p3d_jnt *cur_jntPt, *prev_jntPt;
  configPt q_max_param, q_param;
  double max_param, min_param;
  double range_param = localpathPt->range_param;
  p3d_stay_within_dist_data * stay_within_dist_data;

  return p3d_get_env_dmax()/2;

  /* local path has to be of type quadspline */
  if (localpathPt->type_lp != QUADSPLINE){
    printf("ERROR in p3d_quadspline_stay_within_dist : ");
    printf("Quadspline local path expected\n");
    return 0;
  }
  //on recupere les donnees du localpath
  quadspline_localpathPt = localpathPt->specific.quadspline_data;

  /* store the data to compute the maximal velocities at the
     joint for each body of the robot */
  stay_within_dist_data = MY_ALLOC(p3d_stay_within_dist_data, njnt+2);
  p3d_init_stay_within_dist_data(stay_within_dist_data);

  if (dir == FORWARD){
    q_max_param = quadspline_localpathPt->q_end;
    min_param = max_param = range_param - parameter;
  }
  else{
    q_max_param = quadspline_localpathPt->q_init;
    min_param = max_param = parameter;
  }
  /* Get the current config to have the modifications of the constraints */
  /* Supose that q_init and q_goal respect constraints */
  //q_param = p3d_get_robot_config(robotPt);
  q_param = localpathPt->config_at_param(robotPt,localpathPt,parameter);

  /* computation of the bounds for the linear and angular
     velocities of each body */

  int minJnt = 0;
  for(i=0; i<=njnt; i++) {
    cur_jntPt = robotPt->joints[i];
    prev_jntPt = cur_jntPt->prev_jnt;

    /* j = index of the joint to which the current joint is attached */
    if (prev_jntPt==NULL)
      { j = -1; } /* environment */
    else
      { j = prev_jntPt->num; }
    double bakMinParam = min_param;

    p3d_jnt_stay_within_dist(&(stay_within_dist_data[j+1]), cur_jntPt,
                 &(stay_within_dist_data[i+1]), &(distances[i]),
                 q_param, q_max_param, max_param, &min_param);
    if (min_param < bakMinParam){
      minJnt = cur_jntPt->num;
    }
  /* Rem: stay_within_dist_data[0] is bound to the environment */
  }

  MY_FREE(stay_within_dist_data, p3d_stay_within_dist_data, njnt+2);
  p3d_destroy_config(robotPt, q_param);

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [END] p3d_quadspline_stay_within_dist\n");
    }

  return min_param;
}

p3d_localpath *p3d_quadspline_localplanner( p3d_rob *robotPt, configPt qi, configPt qf, int * ikSol){

    if(DEBUG_LEVEL>0){
        printf("ALEX_DEBUG : [START] p3d_quadspline_local_planner(with speed)\n");
    }

    configPt initconfPt, goalconfPt;
    p3d_localpath *localpathPt;

    /* on verifie que les configurations de depart et d'arrivee existent */
    if(qi == NULL){
        PrintInfo((("MP: p3d_quadspline_localplanner: no start configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_START);
        return(NULL);
    }
    if(qf == NULL){
        PrintInfo((("MP: p3d_quadspline_localplanner: no goal configuration...\n")));
        p3d_set_search_status(P3D_ILLEGAL_GOAL);
        return(NULL);
    }

    /* copy of qi into initconfPt and of qf into goalconfPt */
    initconfPt = p3d_copy_config(robotPt, qi);
    goalconfPt = p3d_copy_config(robotPt, qf);
    if(p3d_get_search_verbose()){
      PrintInfo(("MP: p3d_quadspline_localplanner : "));
      PrintInfo(("qi=("));
      //print_config(robotPt, initconfPt);
      PrintInfo((") ; "));
      PrintInfo(("qf=("));
      //print_config(robotPt, goalconfPt);
      PrintInfo((")\n"));
    }

    /* If initconfPt == goalconfPt, free initconfPt and goalconfPt
     and return NULL */
    if(p3d_equal_config(robotPt,initconfPt, goalconfPt)) {
        PrintInfo((("MP: p3d_quadspline_localplanner: q_init = q_goal!\n")));
        p3d_set_search_status(P3D_CONFIG_EQUAL);
        p3d_destroy_config(robotPt, initconfPt);
        p3d_destroy_config(robotPt, goalconfPt);
        return(NULL);
    }

    localpathPt = p3d_alloc_quadspline_localpath(robotPt,initconfPt,goalconfPt,0,TRUE);

    p3d_set_search_status(P3D_SUCCESS);

    localpathPt->ikSol = ikSol;
    localpathPt->activeCntrts = p3d_getActiveCntrts(robotPt,&(localpathPt->nbActiveCntrts));

    #if defined(LIGHT_PLANNER)
      localpathPt->isCarryingObject = robotPt->isCarryingObject;
      for (int i = 0; i < MAX_CARRIED_OBJECTS; i++){
        if(i < robotPt->armManipulationData->size() ){
          localpathPt->carriedObject[i] = (*robotPt->armManipulationData)[i].getCarriedObject(); /*!< pointer to the carried object (obstacle environment or robot body) */
        }
      }
        //p3d_mat4Copy(robotPt->Tgrasp, localpathMg->Tgrasp);
    #endif

        if(DEBUG_LEVEL>0){
            printf("ALEX_DEBUG : [END] p3d_quadspline_local_planner(with speed)\n");
        }

      //print_quadspline(robotPt,localpathPt);
      return(localpathPt);
}

void print_quadspline(p3d_rob *robotPt,p3d_localpath *localpathPt){
    if(localpathPt->type_lp!=QUADSPLINE){
        printf("ERROR in print_quadspline : localpathPt->type_lp!=QUADSPLINE : localpathPt->type_lp=%s\n",array_localplanner_name[localpathPt->type_lp]);
        return;
    }
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in print_quadspline : quadspline_paramPt is NULL\n");
        return;
    }
    p3d_quadspline_data *quadspline_data=NULL;
    if(localpathPt) quadspline_data=localpathPt->specific.quadspline_data;
    if(quadspline_data==NULL){
        printf("ERROR in print_quadspline : quadspline_data is NULL\n");
        return;
    }
    if(quadspline_paramPt){
        printf("Quadspline ----------------------------------------------------\n");

    printf("   vmax=%f;\n",quadspline_paramPt->vmax);
    printf("   amax=%f;\n",quadspline_paramPt->amax);
    printf("   jmax=%f;\n",quadspline_paramPt->jmax);
    printf("   smax=%f;\n",quadspline_paramPt->smax);
    /*
    printf("   aj=%f\n",quadspline_paramPt->jmax*quadspline_paramPt->jmax/quadspline_paramPt->smax);
    printf("   num_joint=%d\n",quadspline_paramPt->num_joint);
    printf("   x_dof=%d\n",quadspline_paramPt->x_dof);
    printf("   y_dof=%d\n",quadspline_paramPt->y_dof);
    printf("   z_dof=%d\n",quadspline_paramPt->z_dof);
    printf("   rx_dof=%d\n",quadspline_paramPt->rx_dof);
    printf("   ry_dof=%d\n",quadspline_paramPt->ry_dof);
    printf("   rz_dof=%d\n",quadspline_paramPt->rz_dof);
    printf("   vx_dof=%d\n",quadspline_paramPt->vx_dof);
    printf("   vy_dof=%d\n",quadspline_paramPt->vy_dof);
    printf("   vz_dof=%d\n",quadspline_paramPt->vz_dof);
    printf("   ax_dof=%d\n",quadspline_paramPt->ax_dof);
    printf("   ay_dof=%d\n",quadspline_paramPt->ay_dof);
    printf("   az_dof=%d\n",quadspline_paramPt->az_dof);
    printf("   m=%f\n",quadspline_paramPt->mass);
    printf("   f=%f\n",quadspline_paramPt->friction);
    printf("   g=%f\n",quadspline_paramPt->gravity);
    */
    if(quadspline_data){
    printf("\n");
    printf("x0=[%f ",quadspline_data->q_init[quadspline_paramPt->x_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->y_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->z_dof]);
    printf("%f];\n",quadspline_data->q_init[quadspline_paramPt->rz_dof]);
    /*
    printf("r0=[%f ",quadspline_data->q_init[quadspline_paramPt->rx_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->ry_dof]);
    printf("%f];\n",quadspline_data->q_init[quadspline_paramPt->rz_dof]);
    */
    printf("v0=[%f ",quadspline_data->q_init[quadspline_paramPt->vx_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->vy_dof]);
    printf("%f];\n",quadspline_data->q_init[quadspline_paramPt->vz_dof]);
    printf("a0=[%f ",quadspline_data->q_init[quadspline_paramPt->ax_dof]);
    printf("%f ",quadspline_data->q_init[quadspline_paramPt->ay_dof]);
    printf("%f];\n",quadspline_data->q_init[quadspline_paramPt->az_dof]);
    printf("\n");
    printf("xF=[%f ",quadspline_data->q_end[quadspline_paramPt->x_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->y_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->z_dof]);
    printf("%f];\n",quadspline_data->q_end[quadspline_paramPt->rz_dof]);
    /*
    printf("rF=[%f ",quadspline_data->q_end[quadspline_paramPt->rx_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->ry_dof]);
    printf("%f];\n",quadspline_data->q_end[quadspline_paramPt->rz_dof]);
    */
    printf("vF=[%f ",quadspline_data->q_end[quadspline_paramPt->vx_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->vy_dof]);
    printf("%f]\n",quadspline_data->q_end[quadspline_paramPt->vz_dof]);
    printf("aF=[%f ",quadspline_data->q_end[quadspline_paramPt->ax_dof]);
    printf("%f ",quadspline_data->q_end[quadspline_paramPt->ay_dof]);
    printf("%f];\n",quadspline_data->q_end[quadspline_paramPt->az_dof]);
    //if(quadspline_data->calculated){
        printf("\n   signs=[ ");
        for(int k=0;k<4;k++){
            for(int i=0;i<4;i++){
                printf("%d ",(int)quadspline_data->signs[i][k]);
            }
            printf("\n           ");
        }
        printf("];\n");
        printf("\n   durations=[ ");
        for(int k=0;k<4;k++){
            for(int i=0;i<12;i++){
                printf("%f ",quadspline_data->durations[i][k]);
            }
            printf("\n               ");
        }
        printf("];\n");
        //printf("\n   tf=%f\n\n",quadspline_data->tf);
    //}
    }
    /*
    for(int k=0;k<4;k++){
        print_int_ac(quadspline_data,k);
        print_int_eh(quadspline_data,k);
    }
    */
    printf("---------------------------------------------------------------\n");
    }
}

void print_lin(p3d_rob *robotPt,p3d_localpath *localpathPt){
    if(localpathPt->type_lp!=LINEAR){
        printf("ERROR in print_lin : localpathPt->type_lp!=LINEAR : localpathPt->type_lp=%s\n",array_localplanner_name[localpathPt->type_lp]);
        return;
    }
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(quadspline_paramPt==NULL){
        printf("ERROR in print_lin : quadspline_paramPt is NULL\n");
        return;
    }
    p3d_lin_data *lin_data=NULL;
    if(localpathPt) lin_data=localpathPt->specific.lin_data;
    if(lin_data==NULL){
        printf("ERROR in print_lin : lin_data is NULL\n");
        return;
    }
    if(quadspline_paramPt){
        printf("Linear ----------------------------------------------------\n");
        if(lin_data){
            printf("\n");
            printf("q0=[%f ",lin_data->q_init[quadspline_paramPt->x_dof]);
            printf("%f ",lin_data->q_init[quadspline_paramPt->y_dof]);
            printf("%f ",lin_data->q_init[quadspline_paramPt->z_dof]);
            printf("%f];\n",lin_data->q_init[quadspline_paramPt->rz_dof]);
            printf("qF=[%f ",lin_data->q_end[quadspline_paramPt->x_dof]);
            printf("%f ",lin_data->q_end[quadspline_paramPt->y_dof]);
            printf("%f ",lin_data->q_end[quadspline_paramPt->z_dof]);
            printf("%f];\n",lin_data->q_end[quadspline_paramPt->rz_dof]);
        }
        printf("---------------------------------------------------------------\n");
    }
}

void p3d_switch_to_quad(p3d_rob* robotPt,p3d_traj* traj){
    if(!robotPt){
        printf("ERROR in p3d_switch_to_quad : robotPt is NULL\n");
        return;
    }
    p3d_set_robot_steering_method("Quadspline");
    if(!traj){
        printf("ERROR in p3d_switch_to_quad : traj is NULL\n");
        return;
    }
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(!quadspline_paramPt){
        printf("ERROR in p3d_switch_to_quad : quadspline_paramPt is NULL\n");
        return;
    }
    p3d_localpath *localpathPt=traj->courbePt;
    p3d_localpath *prevLP=NULL;
    p3d_localpath *newLP;
    p3d_lin_data *lin_specificPt;
    configPt qi,qf;
    while(localpathPt!=NULL){
        if(localpathPt->type_lp!=LINEAR){
            printf("ERROR in p3d_switch_to_quad : this is not a linear path\n");
            return;
        }
        lin_specificPt=localpathPt->specific.lin_data;
        if(!lin_specificPt){
            printf("ERROR in p3d_switch_to_quad : lin_specificPt is NULL\n");
            return;
        }
        qi=p3d_copy_config(robotPt,lin_specificPt->q_init);
        qf=p3d_copy_config(robotPt,lin_specificPt->q_end);
        // Setting velocity and acceleration to zero
        qi[quadspline_paramPt->vx_dof]=0;
        qi[quadspline_paramPt->vy_dof]=0;
        qi[quadspline_paramPt->vz_dof]=0;
        qi[quadspline_paramPt->ax_dof]=0;
        qi[quadspline_paramPt->ay_dof]=0;
        qi[quadspline_paramPt->az_dof]=0;
        qf[quadspline_paramPt->vx_dof]=0;
        qf[quadspline_paramPt->vy_dof]=0;
        qf[quadspline_paramPt->vz_dof]=0;
        qf[quadspline_paramPt->ax_dof]=0;
        qf[quadspline_paramPt->ay_dof]=0;
        qf[quadspline_paramPt->az_dof]=0;
        newLP=p3d_quadspline_localplanner(robotPt,qi,qf,robotPt->ikSol);
        if(!newLP){
            printf("ERROR in p3d_switch_to_quad : newLP is NULL\n");
            return;
        }
        if(!prevLP){
            traj->courbePt=newLP;
            traj->range_param=newLP->range_param;
        }
        else{
            prevLP->next_lp=newLP;
            traj->range_param=traj->range_param+newLP->range_param;
        }
        newLP->prev_lp=prevLP;
        newLP->next_lp=localpathPt->next_lp;
        localpathPt->destroy(robotPt,localpathPt);
        localpathPt=newLP->next_lp;
        prevLP=newLP;
    }
}

void p3d_switch_to_quad_from_FF(p3d_rob* robotPt,p3d_traj* traj){
    if(!traj){
        printf("ERROR in p3d_switch_to_quad_from_FF : traj is NULL\n");
        return;
    }
    if(!robotPt){
        printf("ERROR in p3d_switch_to_quad_from_FF : robotPt is NULL\n");
        return;
    }
    p3d_set_robot_steering_method("Quadspline");
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(!quadspline_paramPt){
        printf("ERROR in p3d_switch_to_quad_from_FF : quadspline_paramPt is NULL\n");
        return;
    }
    p3d_localpath *localpathPt=traj->courbePt;
    p3d_localpath *prevLP=NULL;
    p3d_localpath *newLP;
    p3d_lin_data *lin_specificPt;
    configPt qi,qf;
    while(localpathPt!=NULL){
        if(localpathPt->type_lp!=LINEAR){
            printf("ERROR in p3d_switch_to_quad_from_FF : this is not a linear path\n");
            return;
        }
        lin_specificPt=localpathPt->specific.lin_data;
        if(!lin_specificPt){
            printf("ERROR in p3d_switch_to_quad_from_FF : lin_specificPt is NULL\n");
            return;
        }

        qi=p3d_copy_config(robotPt,robotPt->ROBOT_POS);
        qf=p3d_copy_config(robotPt,robotPt->ROBOT_POS);
        // Coping relevant data and setting velocity and acceleration to zero
        for(int k=0;k<6;k++){
            qi[quadspline_paramPt->x_dof+k]=lin_specificPt->q_init[k+6];
            qf[quadspline_paramPt->x_dof+k]=lin_specificPt->q_end[k+6];
            qi[quadspline_paramPt->vx_dof+k]=0.0;
            qf[quadspline_paramPt->vx_dof+k]=0.0;
        }

        newLP=p3d_quadspline_localplanner(robotPt,qi,qf,robotPt->ikSol);
        if(!newLP){
            printf("ERROR in p3d_switch_to_quad_from_FF : newLP is NULL\n");
            return;
        }
        if(!prevLP){
            traj->courbePt=newLP;
            traj->range_param=newLP->range_param;
        }
        else{
            prevLP->next_lp=newLP;
            traj->range_param=traj->range_param+newLP->range_param;
        }
        newLP->prev_lp=prevLP;
        newLP->next_lp=localpathPt->next_lp;
        localpathPt->destroy(traj->rob,localpathPt);
        localpathPt=newLP->next_lp;
        prevLP=newLP;
    }
    traj->rob=robotPt;
    printf("Done (p3d_switch_to_quad_from_FF)\n");
}

void p3d_switch_to_lin(p3d_rob* robotPt,p3d_traj* traj){
    if(!robotPt){
        printf("ERROR in p3d_switch_to_lin : robotPt is NULL\n");
        return;
    }
    p3d_set_robot_steering_method("Linear");
    if(!traj){
        printf("ERROR in p3d_switch_to_lin : traj is NULL\n");
        return;
    }
    lm_quadspline_str *quadspline_paramPt=lm_get_quadspline_lm_param(robotPt);
    if(!quadspline_paramPt){
        printf("ERROR in p3d_switch_to_lin : quadspline_paramPt is NULL\n");
        return;
    }
    p3d_localpath *localpathPt=traj->courbePt;
    p3d_localpath *prevLP=NULL;
    p3d_localpath *newLP;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    while(localpathPt!=NULL){
        if(localpathPt->type_lp!=QUADSPLINE){
            printf("ERROR in p3d_switch_to_lin : this is not a Quadspline path\n");
            return;
        }
        quadspline_specificPt=localpathPt->specific.quadspline_data;
        if(!quadspline_specificPt){
            printf("ERROR in p3d_switch_to_lin : quadspline_specificPt is NULL\n");
            return;
        }
        qi=p3d_copy_config(robotPt,quadspline_specificPt->q_init);
        qf=p3d_copy_config(robotPt,quadspline_specificPt->q_end);
        // Setting velocity and acceleration to zero
        qi[quadspline_paramPt->vx_dof]=0;
        qi[quadspline_paramPt->vy_dof]=0;
        qi[quadspline_paramPt->vz_dof]=0;
        qi[quadspline_paramPt->ax_dof]=0;
        qi[quadspline_paramPt->ay_dof]=0;
        qi[quadspline_paramPt->az_dof]=0;
        qf[quadspline_paramPt->vx_dof]=0;
        qf[quadspline_paramPt->vy_dof]=0;
        qf[quadspline_paramPt->vz_dof]=0;
        qf[quadspline_paramPt->ax_dof]=0;
        qf[quadspline_paramPt->ay_dof]=0;
        qf[quadspline_paramPt->az_dof]=0;
        newLP=p3d_local_planner(robotPt,qi,qf);
        if(!newLP){
            printf("ERROR in p3d_switch_to_quad : newLP is NULL\n");
            return;
        }
        if(!prevLP){
            traj->courbePt=newLP;
            traj->range_param=newLP->range_param;
        }
        else{
            prevLP->next_lp=newLP;
            traj->range_param=traj->range_param+newLP->range_param;
        }
        newLP->prev_lp=prevLP;
        newLP->next_lp=localpathPt->next_lp;
        localpathPt->destroy(robotPt,localpathPt);
        localpathPt=newLP->next_lp;
        prevLP=newLP;
    }
}

void unit_test_case_abc(p3d_rob* robotPt,int nbtest){

    int k,comp;
    double aB;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"case_abc");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);

        aB=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);

        for(comp=0;comp<4;comp++){
            fprintf(f,"case_move3d=%d;\n",case_abc(aB,quadspline_paramPt,quadspline_specificPt,comp));
            fprintf(f,"case_matlab=case_abc(%f,a0(%d),jmax,smax);\n",aB,comp+1);
            fprintf(f,"if(case_matlab~=case_move3d)\n");
            fprintf(f,"   disp([case_matlab case_move3d %f a0(%d) %d]);\n",aB,comp+1,comp+1);
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_case_egh(p3d_rob* robotPt,int nbtest){

    int k,comp;
    double aG;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"case_egh");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);

        aG=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);

        for(comp=0;comp<4;comp++){
            fprintf(f,"case_move3d=%d;\n",case_egh(aG,quadspline_paramPt,quadspline_specificPt,comp));
            fprintf(f,"case_matlab=case_egh(%f,aF(%d),jmax,smax);\n",aG,comp+1);
            fprintf(f,"if(case_matlab~=case_move3d)\n");
            fprintf(f,"   disp([case_matlab case_move3d %f a0(%d) %d]);\n",aG,comp+1,comp+1);
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_durations_and_signs_ac(p3d_rob* robotPt,int nbtest){

    int k,comp;
    double aB;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"durations_and_signs_ac");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        aB=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        for(comp=0;comp<4;comp++){
            durations_and_signs_ac(aB,quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"tA1=%f;\n",quadspline_specificPt->durations[A1][comp]);
            fprintf(f,"tA2=%f;\n",quadspline_specificPt->durations[A2][comp]);
            fprintf(f,"tC1=%f;\n",quadspline_specificPt->durations[C1][comp]);
            fprintf(f,"tC2=%f;\n",quadspline_specificPt->durations[C2][comp]);
            fprintf(f,"sA=%f;\n",quadspline_specificPt->signs[A][comp]);
            fprintf(f,"sC=%f;\n",quadspline_specificPt->signs[C][comp]);
            fprintf(f,"move3d=[tA1,tA2,tC1,tC2,sA,sC];\n");
            fprintf(f,"[tA1,tA2,tC1,tC2,sA,sC]=durations_and_signs_ac(%f,a0(%d),jmax,smax);\n",aB,comp+1);
            fprintf(f,"matlab=[tA1,tA2,tC1,tC2,sA,sC];\n");
            fprintf(f,"if(norm(move3d-matlab)>1e-5)\n");
            fprintf(f,"   disp('----------------------------------------');\n");
            fprintf(f,"   matlab\n");
            fprintf(f,"   move3d\n");
            fprintf(f,"   disp('----------------------------------------');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_durations_and_signs_eh(p3d_rob* robotPt,int nbtest){

    int k,comp;
    double aG;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"durations_and_signs_eh");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        aG=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        for(comp=0;comp<4;comp++){
            durations_and_signs_eh(aG,quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"tE1=%f;\n",quadspline_specificPt->durations[E1][comp]);
            fprintf(f,"tE2=%f;\n",quadspline_specificPt->durations[E2][comp]);
            fprintf(f,"tH1=%f;\n",quadspline_specificPt->durations[H1][comp]);
            fprintf(f,"tH2=%f;\n",quadspline_specificPt->durations[H2][comp]);
            fprintf(f,"sE=%f;\n",quadspline_specificPt->signs[E][comp]);
            fprintf(f,"sH=%f;\n",quadspline_specificPt->signs[H][comp]);
            fprintf(f,"move3d=[tE1,tE2,tH1,tH2,sE,sH];\n");
            fprintf(f,"[tE1,tE2,tH1,tH2,sE,sH]=durations_and_signs_eh(%f,aF(%d),jmax,smax);\n",aG,comp+1);
            fprintf(f,"matlab=[tE1,tE2,tH1,tH2,sE,sH];\n");
            fprintf(f,"if(norm(move3d-matlab)>1e-5)\n");
            fprintf(f,"   disp('----------------------------------------');\n");
            fprintf(f,"   matlab\n");
            fprintf(f,"   move3d\n");
            fprintf(f,"   disp('----------------------------------------');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_x_c(p3d_rob* robotPt,int nbtest){

    int k,comp;
    double aB,tB;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"x_c");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"x0=[%f %f %f %f];\n",qi[quadspline_paramPt->x_dof]
                                       ,qi[quadspline_paramPt->y_dof]
                                       ,qi[quadspline_paramPt->z_dof]
                                       ,qi[quadspline_paramPt->rz_dof]);
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        aB=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        tB=p3d_random(0,1);
        if(tB<0.5){
            tB=0;
        }
        else{
            tB=p3d_random(0,20);
        }
        for(comp=0;comp<4;comp++){
            fprintf(f,"xC_move3d=%f;\n",x_c(aB,tB,quadspline_paramPt,quadspline_specificPt,comp,0));
            fprintf(f,"xC_matlab=x_c(%f,%f,x0(%d),v0(%d),a0(%d),jmax,smax);\n",aB,tB,comp+1,comp+1,comp+1);
            fprintf(f,"if(norm(xC_move3d-xC_matlab)>1e-3)\n");
            fprintf(f,"   disp(xC_move3d-xC_matlab);\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_x_e(p3d_rob* robotPt,int nbtest){
    int k,comp;
    double aG,tG;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"x_e");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"xF=[%f %f %f %f];\n",qf[quadspline_paramPt->x_dof]
                                       ,qf[quadspline_paramPt->y_dof]
                                       ,qf[quadspline_paramPt->z_dof]
                                       ,qf[quadspline_paramPt->rz_dof]);
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        aG=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        tG=p3d_random(0,1);
        if(tG<0.5){
            tG=0;
        }
        else{
            tG=p3d_random(0,20);
        }
        for(comp=0;comp<4;comp++){
            fprintf(f,"xE_move3d=%f;\n",x_e(aG,tG,quadspline_paramPt,quadspline_specificPt,comp,0));
            fprintf(f,"xE_matlab=x_e(%f,%f,xF(%d),vF(%d),aF(%d),jmax,smax);\n",aG,tG,comp+1,comp+1,comp+1);
            fprintf(f,"if(norm(xE_move3d-xE_matlab)>1e-3)\n");
            fprintf(f,"   disp(xE_move3d-xE_matlab);\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_v_c(p3d_rob* robotPt,int nbtest){
    int k,comp;
    double aB,tB;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"v_c");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        aB=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        tB=p3d_random(0,1);
        if(tB<0.5){
            tB=0;
        }
        else{
            tB=p3d_random(0,20);
        }
        for(comp=0;comp<4;comp++){
            fprintf(f,"vC_move3d=%f;\n",v_c(aB,tB,quadspline_paramPt,quadspline_specificPt,comp,0));
            fprintf(f,"vC_matlab=v_c(%f,%f,v0(%d),a0(%d),jmax,smax);\n",aB,tB,comp+1,comp+1,comp+1);
            fprintf(f,"if(norm(vC_move3d-vC_matlab)>1e-3)\n");
            fprintf(f,"   disp(vC_move3d-vC_matlab);\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_v_e(p3d_rob* robotPt,int nbtest){
    int k,comp;
    double aG,tG;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"v_e");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        aG=p3d_random(-quadspline_paramPt->amax,quadspline_paramPt->amax);
        tG=p3d_random(0,1);
        if(tG<0.5){
            tG=0;
        }
        else{
            tG=p3d_random(0,20);
        }
        for(comp=0;comp<4;comp++){
            fprintf(f,"vE_move3d=%f;\n",v_e(aG,tG,quadspline_paramPt,quadspline_specificPt,comp,0));
            fprintf(f,"vE_matlab=v_e(%f,%f,vF(%d),aF(%d),jmax,smax);\n",aG,tG,comp+1,comp+1,comp+1);
            fprintf(f,"if(norm(vE_move3d-vE_matlab)>1e-3)\n");
            fprintf(f,"   disp(vE_move3d-vE_matlab);\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_intervals_ac(p3d_rob* robotPt,int nbtest){
    int k,comp,l;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"intervals_ac");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        for(comp=0;comp<4;comp++){
            intervals_ac(quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"int_ac_move3d=zeros(3,%d);\n",quadspline_specificPt->cases_abc[0][comp]+1);
            for(l=0;l<=quadspline_specificPt->cases_abc[0][comp];l++){
                fprintf(f,"int_ac_move3d(:,%d)=[%d;%f;%f];\n",l+1
                                                           ,quadspline_specificPt->cases_abc[l][comp]
                                                           ,quadspline_specificPt->int_v_abc[l][comp]
                                                           ,quadspline_specificPt->int_a_abc[l][comp]);
            }
            fprintf(f,"int_ac_matlab=intervals_ac(v0(%d),a0(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"if(size(int_ac_matlab,1)~=size(int_ac_move3d,1) || size(int_ac_matlab,2)~=size(int_ac_move3d,2))\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('SIZE (comp=%d)');\n",comp+1);
            fprintf(f,"   disp('--matlab--------------------------------');\n");
            fprintf(f,"   disp(int_ac_matlab);");
            fprintf(f,"   disp('--move3d--------------------------------');\n");
            fprintf(f,"   disp(int_ac_move3d);");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   N=0;\n");
            fprintf(f,"   for k=1:3\n");
            fprintf(f,"      N=N+norm(int_ac_matlab(k,:)-int_ac_move3d(k,:));\n");
            fprintf(f,"   end\n");
            fprintf(f,"   if(N>1e-3)\n");
            fprintf(f,"      disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"      disp('VALUES (comp=%d)');\n",comp+1);
            fprintf(f,"     disp('--matlab--------------------------------');\n");
            fprintf(f,"     disp(int_ac_matlab);");
            fprintf(f,"     disp('--move3d--------------------------------');\n");
            fprintf(f,"     disp(int_ac_move3d);");
            fprintf(f,"     disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"      nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"   else\n");
            fprintf(f,"      nb_valid=nb_valid+1;\n");
            fprintf(f,"   end\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_intervals_eh(p3d_rob* robotPt,int nbtest){
    int k,comp,l;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"intervals_eh");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"format shortg\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        for(comp=0;comp<4;comp++){
            intervals_eh(quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"int_eh_move3d=zeros(3,%d);\n",quadspline_specificPt->cases_egh[0][comp]+1);
            for(l=0;l<=quadspline_specificPt->cases_egh[0][comp];l++){
                fprintf(f,"int_eh_move3d(:,%d)=[%d;%f;%f];\n",l+1
                                                           ,quadspline_specificPt->cases_egh[l][comp]
                                                           ,quadspline_specificPt->int_v_egh[l][comp]
                                                           ,quadspline_specificPt->int_a_egh[l][comp]);
            }
            fprintf(f,"int_eh_matlab=intervals_eh(vF(%d),aF(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"if(size(int_eh_matlab,1)~=size(int_eh_move3d,1) || size(int_eh_matlab,2)~=size(int_eh_move3d,2))\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('SIZE (comp=%d)');\n",comp+1);
            fprintf(f,"   disp('--matlab--------------------------------');\n");
            fprintf(f,"   disp(int_eh_matlab);");
            fprintf(f,"   disp('--move3d--------------------------------');\n");
            fprintf(f,"   disp(int_eh_move3d);");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   N=0;\n");
            fprintf(f,"   for k=1:3\n");
            fprintf(f,"      N=N+norm(int_eh_matlab(k,:)-int_eh_move3d(k,:));\n");
            fprintf(f,"   end\n");
            fprintf(f,"   if(N>1e-3)\n");
            fprintf(f,"      disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"      disp('VALUES (comp=%d)');\n",comp+1);
            fprintf(f,"     disp('--matlab--------------------------------');\n");
            fprintf(f,"     disp(int_eh_matlab);");
            fprintf(f,"     disp('--move3d--------------------------------');\n");
            fprintf(f,"     disp(int_eh_move3d);");
            fprintf(f,"     disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"      nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"   else\n");
            fprintf(f,"      nb_valid=nb_valid+1;\n");
            fprintf(f,"   end\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_a_b(p3d_rob* robotPt,int nbtest){
    int k,comp;
    double vC;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"a_b");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);
    fprintf(f,"epsi=%f;\n",EPSI);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        vC=p3d_random(-quadspline_paramPt->vmax,quadspline_paramPt->vmax);
        for(comp=0;comp<4;comp++){
            intervals_ac(quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"aB_move3d=%f;\n",a_b(vC,quadspline_paramPt,quadspline_specificPt,comp));
            fprintf(f,"tB_move3d=%f;\n",quadspline_specificPt->durations[B][comp]);
            fprintf(f,"int_ac=intervals_ac(v0(%d),a0(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"[aB_matlab,tB_matlab]=a_b(%f,v0(%d),a0(%d),amax,jmax,smax,int_ac,epsi);\n",vC,comp+1,comp+1);
            fprintf(f,"if(norm(aB_matlab-aB_move3d)>1e-3 || norm(tB_matlab-tB_move3d)>1e-3)\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('comp=%d');\n",comp+1);
            fprintf(f,"   disp('--matlab-[aB,tB]------------------------');\n");
            fprintf(f,"   disp([aB_matlab tB_matlab]);");
            fprintf(f,"   disp('--move3d-[aB,tB]------------------------');\n");
            fprintf(f,"   disp([aB_move3d tB_move3d]);");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_a_g(p3d_rob* robotPt,int nbtest){
    int k,comp;
    double vE;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"a_g");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);
    fprintf(f,"epsi=%f;\n",EPSI);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        vE=p3d_random(-quadspline_paramPt->vmax,quadspline_paramPt->vmax);
        for(comp=0;comp<4;comp++){
            intervals_eh(quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"aG_move3d=%f;\n",a_g(vE,quadspline_paramPt,quadspline_specificPt,comp));
            fprintf(f,"tG_move3d=%f;\n",quadspline_specificPt->durations[G][comp]);
            fprintf(f,"int_eh=intervals_eh(vF(%d),aF(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"[aG_matlab,tG_matlab]=a_g(%f,vF(%d),aF(%d),amax,jmax,smax,int_eh,epsi);\n",vE,comp+1,comp+1);
            fprintf(f,"if(norm(aG_matlab-aG_move3d)>1e-3 || norm(tG_matlab-tG_move3d)>1e-3)\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('comp=%d');\n",comp+1);
            fprintf(f,"   disp('--matlab-[aG,tG]------------------------');\n");
            fprintf(f,"   disp([aG_matlab tG_matlab]);");
            fprintf(f,"   disp('--move3d-[aB,tB]------------------------');\n");
            fprintf(f,"   disp([aG_move3d tG_move3d]);");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_d_x(p3d_rob* robotPt,int nbtest){
    int k,comp,l;
    double vD,dX;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"d_x");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);
    fprintf(f,"epsi=%f;\n",EPSI);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"x0=[%f %f %f %f];\n",qi[quadspline_paramPt->x_dof]
                                       ,qi[quadspline_paramPt->y_dof]
                                       ,qi[quadspline_paramPt->z_dof]
                                       ,qi[quadspline_paramPt->rz_dof]);
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        fprintf(f,"xF=[%f %f %f %f];\n",qf[quadspline_paramPt->x_dof]
                                       ,qf[quadspline_paramPt->y_dof]
                                       ,qf[quadspline_paramPt->z_dof]
                                       ,qf[quadspline_paramPt->rz_dof]);
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        vD=p3d_random(-quadspline_paramPt->vmax,quadspline_paramPt->vmax);
        for(comp=0;comp<4;comp++){
            intervals_ac(quadspline_paramPt,quadspline_specificPt,comp);
            intervals_eh(quadspline_paramPt,quadspline_specificPt,comp);
            dX=d_x(vD,quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"move3d=[%f",dX);
            for(l=0;l<12;l++){
                fprintf(f,",%f",quadspline_specificPt->durations[l][comp]);
            }
            for(l=0;l<4;l++){
                fprintf(f,",%f",quadspline_specificPt->signs[l][comp]);
            }
            fprintf(f,"];\n");
            fprintf(f,"int_ac=intervals_ac(v0(%d),a0(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"int_eh=intervals_eh(vF(%d),aF(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"[dX,tF,durations,signs]=d_x(%f,x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),amax,jmax,smax,int_ac,int_eh,epsi);\n",vD,comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"matlab=[dX durations tF signs];\n");
            fprintf(f,"if(norm(matlab-move3d)/norm(matlab)>1e-3)\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('comp=%d');\n",comp+1);
            fprintf(f,"   disp(norm(matlab-move3d));\n");
            fprintf(f,"   disp('--[matlab move3d]-----------------------');\n");
            fprintf(f,"   disp([matlab' move3d']);\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}
void unit_test_v_optim(p3d_rob* robotPt,int nbtest){
    int k,comp,l;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"v_optim");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);
    fprintf(f,"epsi=%f;\n",EPSI);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"x0=[%f %f %f %f];\n",qi[quadspline_paramPt->x_dof]
                                       ,qi[quadspline_paramPt->y_dof]
                                       ,qi[quadspline_paramPt->z_dof]
                                       ,qi[quadspline_paramPt->rz_dof]);
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        fprintf(f,"xF=[%f %f %f %f];\n",qf[quadspline_paramPt->x_dof]
                                       ,qf[quadspline_paramPt->y_dof]
                                       ,qf[quadspline_paramPt->z_dof]
                                       ,qf[quadspline_paramPt->rz_dof]);
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        for(comp=0;comp<4;comp++){
            v_optim(quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"move3d=[%f",quadspline_specificPt->v_opt[comp]);
            for(l=0;l<12;l++){
                fprintf(f,",%f",quadspline_specificPt->durations[l][comp]);
            }
            for(l=0;l<4;l++){
                fprintf(f,",%f",quadspline_specificPt->signs[l][comp]);
            }
            fprintf(f,"];\n");
            fprintf(f,"int_ac=intervals_ac(v0(%d),a0(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"int_eh=intervals_eh(vF(%d),aF(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"[v,tF,durations,signs]=v_opt(x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),vmax,amax,jmax,smax,int_ac,int_eh,epsi);\n",comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"matlab=[v durations tF signs];\n");
            fprintf(f,"if(abs(matlab(1)-move3d(1))>1e-3 || norm(matlab-move3d)/norm(matlab)>1e-3)\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('comp=%d');\n",comp+1);
            fprintf(f,"   disp(norm(matlab-move3d));\n");
            fprintf(f,"   disp('--[matlab move3d]-----------------------');\n");
            fprintf(f,"   disp([matlab' move3d']);\n");
            fprintf(f,"   Tmax=50;\n");
            fprintf(f,"   N=100;\n");
            fprintf(f,"   V=[-vmax:2*vmax/(N-1):vmax];\n");
            fprintf(f,"   dX=zeros(1,N);\n");
            fprintf(f,"   for k=1:N\n");
            fprintf(f,"      dX(k)=d_x(V(k),x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),amax,jmax,smax,int_ac,int_eh,epsi);\n",comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"   end\n");
            fprintf(f,"   dXmax=max(dX);\n");
            fprintf(f,"   dXmin=min(dX);\n");
            fprintf(f,"   close all\n");
            fprintf(f,"   figure\n");
            fprintf(f,"   hold on\n");
            fprintf(f,"   plot([-vmax vmax],[0 0],'k');\n");
            fprintf(f,"   plot([0 0],[dXmin dXmax],'k');\n");
            fprintf(f,"   plot(V,dX,'b');\n");
            fprintf(f,"   dX_matlab=d_x(matlab(1),x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),amax,jmax,smax,int_ac,int_eh,epsi);\n",comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"   dX_move3d=d_x(move3d(1),x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),amax,jmax,smax,int_ac,int_eh,epsi);\n",comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"   plot(matlab(1),dX_matlab,'rs')\n");
            fprintf(f,"   plot(move3d(1),dX_move3d,'gs')\n");
            fprintf(f,"   axis([-1.1*vmax 1.1*vmax dXmin-0.1*(dXmax-dXmin) dXmax+0.1*(dXmax-dXmin)]);\n");
            fprintf(f,"   drawnow\n");
            fprintf(f,"   figure\n");
            fprintf(f,"   hold on\n");
            fprintf(f,"   durations=matlab(2:12);\n");
            fprintf(f,"   signs=matlab(14:17);\n");
            fprintf(f,"   [T,X,V,A,Je]=spline_vector(x0,v0,a0,xF,vF,aF,smax,durations,signs,1000);\n");
            fprintf(f,"   disp_spline(T,X,V,A,Je,vmax,amax,jmax,durations,signs);\n");
            fprintf(f,"   drawnow\n");
            fprintf(f,"   figure\n");
            fprintf(f,"   hold on\n");
            fprintf(f,"   durations=move3d(2:12);\n");
            fprintf(f,"   signs=move3d(14:17);\n");
            fprintf(f,"   [T,X,V,A,Je]=spline_vector(x0,v0,a0,xF,vF,aF,smax,durations,signs,1000);\n");
            fprintf(f,"   disp_spline(T,X,V,A,Je,vmax,amax,jmax,durations,signs);\n");
            fprintf(f,"   drawnow\n");
            fprintf(f,"   pause\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }
    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_synchronize(p3d_rob* robotPt,int nbtest){
    int k,comp,l,slowest_comp,ERROR;
    double tF_des;
    lm_quadspline_str *quadspline_paramPt;
    p3d_quadspline_data *quadspline_specificPt;
    configPt qi,qf;
    FILE *f;
    char function_name[100];
    char file_name[100];
    strcpy(function_name,"unit_test_");
    strcat(function_name,"synchronize");
    strcpy(file_name,function_name);
    strcat(file_name,".m");

    printf("   %s ... ",function_name);

    if(!robotPt){
        printf("\nERROR in %s : robotPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_paramPt=lm_get_quadspline_lm_param(robotPt))){
        printf("\nERROR in %s : quadspline_paramPt is NULL\n",function_name);
        return;
    }
    if(!(quadspline_specificPt=MY_ALLOC(p3d_quadspline_data,1))){
        printf("\nERROR in %s : can not allocate quadspline_specificPt\n",function_name);
        return;
    }
    if(!(qi=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qi\n",function_name);
        return;
    }
    quadspline_specificPt->q_init=qi;
    if(!(qf=p3d_alloc_config(robotPt))){
        printf("\nERROR in %s : can not allocate qf\n",function_name);
        return;
    }
    quadspline_specificPt->q_end=qf;
    if(!(f=fopen(file_name,"w"))){
        printf("\nERROR in %s : can not open file %s\n",function_name,file_name);
        return;
    }
    fprintf(f,"%c %s generated by move3d\n",'%',function_name);
    fprintf(f,"clear all\n");
    fprintf(f,"nb_valid=0;\n");
    fprintf(f,"nb_unvalid=0;\n");
    fprintf(f,"vmax=%f;\n",quadspline_paramPt->vmax);
    fprintf(f,"amax=%f;\n",quadspline_paramPt->amax);
    fprintf(f,"jmax=%f;\n",quadspline_paramPt->jmax);
    fprintf(f,"smax=%f;\n",quadspline_paramPt->smax);
    fprintf(f,"epsi=%f;\n",EPSI);

    for(k=0;k<nbtest;k++){
        p3d_shoot(robotPt,qi,0);
        p3d_shoot(robotPt,qf,0);
        fprintf(f,"\n");
        fprintf(f,"x0=[%f %f %f %f];\n",qi[quadspline_paramPt->x_dof]
                                       ,qi[quadspline_paramPt->y_dof]
                                       ,qi[quadspline_paramPt->z_dof]
                                       ,qi[quadspline_paramPt->rz_dof]);
        fprintf(f,"v0=[%f %f %f 0];\n",qi[quadspline_paramPt->vx_dof]
                                      ,qi[quadspline_paramPt->vy_dof]
                                      ,qi[quadspline_paramPt->vz_dof]);
        fprintf(f,"a0=[%f %f %f 0];\n",qi[quadspline_paramPt->ax_dof]
                                      ,qi[quadspline_paramPt->ay_dof]
                                      ,qi[quadspline_paramPt->az_dof]);
        fprintf(f,"xF=[%f %f %f %f];\n",qf[quadspline_paramPt->x_dof]
                                       ,qf[quadspline_paramPt->y_dof]
                                       ,qf[quadspline_paramPt->z_dof]
                                       ,qf[quadspline_paramPt->rz_dof]);
        fprintf(f,"vF=[%f %f %f 0];\n",qf[quadspline_paramPt->vx_dof]
                                      ,qf[quadspline_paramPt->vy_dof]
                                      ,qf[quadspline_paramPt->vz_dof]);
        fprintf(f,"aF=[%f %f %f 0];\n",qf[quadspline_paramPt->ax_dof]
                                      ,qf[quadspline_paramPt->ay_dof]
                                      ,qf[quadspline_paramPt->az_dof]);
        tF_des=0;
        for(comp=0;comp<4;comp++){
            v_optim(quadspline_paramPt,quadspline_specificPt,comp);
            if(quadspline_specificPt->durations[F][comp]>tF_des){
                tF_des=quadspline_specificPt->durations[F][comp];
                slowest_comp=comp;
            }
        }

        fprintf(f,"tF_des_move3d=%f;\n",tF_des);

        fprintf(f,"tF_des_matlab=0;\n");
        fprintf(f,"tF=zeros(1,4);\n");
        fprintf(f,"vD=zeros(1,4);\n");
        fprintf(f,"for k=1:4\n");
        fprintf(f,"   int_ac=intervals_ac(v0(k),a0(k),amax,jmax,smax);\n");
        fprintf(f,"   int_eh=intervals_eh(vF(k),aF(k),amax,jmax,smax);\n");
        fprintf(f,"   [vD(k),tF(k)]=v_opt(x0(k),v0(k),a0(k),xF(k),vF(k),aF(k),vmax,amax,jmax,smax,int_ac,int_eh,epsi);\n");
        fprintf(f,"   if(tF(k)>tF_des_matlab)\n");
        fprintf(f,"      tF_des_matlab=tF(k);\n");
        fprintf(f,"   end\n");
        fprintf(f,"end\n");

        ERROR=0;
        for(comp=0;comp<4;comp++){
            ERROR=(ERROR || (fabs(quadspline_specificPt->v_opt[comp])>quadspline_paramPt->vmax));
        }

        for(comp=0;comp<4;comp++){
            if(ERROR) v_optim(quadspline_paramPt,quadspline_specificPt,comp);
            synchronize(tF_des,quadspline_paramPt,quadspline_specificPt,comp);
            fprintf(f,"move3d=[%f",quadspline_specificPt->durations[0][comp]);
            for(l=1;l<12;l++){
                fprintf(f,",%f",quadspline_specificPt->durations[l][comp]);
            }
            for(l=0;l<4;l++){
                fprintf(f,",%f",quadspline_specificPt->signs[l][comp]);
            }
            fprintf(f,"];\n");
            fprintf(f,"int_ac_move3d=zeros(3,%d);\n",quadspline_specificPt->cases_abc[0][comp]+1);
            for(l=0;l<=quadspline_specificPt->cases_abc[0][comp];l++){
                fprintf(f,"int_ac_move3d(:,%d)=[%d;%f;%f];\n",l+1
                                                             ,quadspline_specificPt->cases_abc[l][comp]
                                                             ,quadspline_specificPt->int_v_abc[l][comp]
                                                             ,quadspline_specificPt->int_a_abc[l][comp]);
            }
            fprintf(f,"int_ac=intervals_ac(v0(%d),a0(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"int_eh_move3d=zeros(3,%d);\n",quadspline_specificPt->cases_egh[0][comp]+1);
            for(l=0;l<=quadspline_specificPt->cases_egh[0][comp];l++){
                fprintf(f,"int_eh_move3d(:,%d)=[%d;%f;%f];\n",l+1
                                                             ,quadspline_specificPt->cases_egh[l][comp]
                                                             ,quadspline_specificPt->int_v_egh[l][comp]
                                                             ,quadspline_specificPt->int_a_egh[l][comp]);
            }
            fprintf(f,"int_eh=intervals_eh(vF(%d),aF(%d),amax,jmax,smax);\n",comp+1,comp+1);
            fprintf(f,"[v,t,durations,signs]=synchronize(tF_des_matlab,tF(%d),vD(%d),x0(%d),v0(%d),a0(%d),xF(%d),vF(%d),aF(%d),amax,jmax,smax,int_ac,int_eh,epsi);\n",comp+1,comp+1,comp+1,comp+1,comp+1,comp+1,comp+1,comp+1);
            fprintf(f,"matlab=[durations t signs];\n");
            fprintf(f,"if(abs(tF_des_matlab-tF_des_move3d)>1e-3 || abs(tF_des_matlab-t)>1e-3 || norm(matlab-move3d)/norm(matlab)>1e-3)\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   disp('comp=%d');\n",comp+1);
            fprintf(f,"   disp(norm(matlab-move3d));\n");
            fprintf(f,"   disp('--[tF_des_matlab tF_des_move3d]-----');\n");
            fprintf(f,"   disp([tF_des_matlab tF_des_move3d]);\n");
            fprintf(f,"   disp('--[int_ac_matlab-int_ac_move3d]-----');\n");
            fprintf(f,"   disp(int_ac-int_ac_move3d);\n");
            fprintf(f,"   disp('--[int_eh_matlab-int_eh_move3d]-----');\n");
            fprintf(f,"   disp(int_eh-int_eh_move3d);\n");
            fprintf(f,"   disp('--[matlab move3d]-----------------------');\n");
            fprintf(f,"   disp([matlab' move3d']);\n");
            fprintf(f,"   disp('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>');\n");
            fprintf(f,"   nb_unvalid=nb_unvalid+1;\n");
            fprintf(f,"else\n");
            fprintf(f,"   nb_valid=nb_valid+1;\n");
            fprintf(f,"end\n");
        }

    }

    fprintf(f,"if(nb_valid==%d && nb_unvalid==0)\n",4*nbtest);
    fprintf(f,"   disp('%s : OK');",function_name);
    fprintf(f,"else\n");
    fprintf(f,"   disp('%s : KO (%d)');",function_name,4*nbtest);
    fprintf(f,"   disp(nb_unvalid);");
    fprintf(f,"end\n");

    fclose(f);
    p3d_destroy_config(robotPt,qi);
    p3d_destroy_config(robotPt,qf);
    MY_FREE(quadspline_specificPt,p3d_quadspline_data,1);
    printf("done\n");
}

void unit_test_spline_at_from_q0(p3d_rob* robotPt,int nbtest){

}
void unit_test_spline_at_from_qF(p3d_rob* robotPt,int nbtest){

}

void unit_test_spline_at(p3d_rob* robotPt,int nbtest){

}

void unit_test(p3d_rob* robotPt,int nbtest){
    FILE *f;
    printf("\nunit_test ---------------------------------------------------\n");
    if(!robotPt){
        printf("\nERROR in unit_test : robotPt is NULL\n");
        return;
    }
    if(!(f=fopen("unit_test.m","w"))){
        printf("\nERROR in unit_test : can not open file unit_test.m\n");
        return;
    }
    unit_test_case_abc(robotPt,nbtest);
    fprintf(f,"unit_test_case_abc\n");
    unit_test_case_egh(robotPt,nbtest);
    fprintf(f,"unit_test_case_egh\n");
    unit_test_durations_and_signs_ac(robotPt,nbtest);
    fprintf(f,"unit_test_durations_and_signs_ac\n");
    unit_test_durations_and_signs_eh(robotPt,nbtest);
    fprintf(f,"unit_test_durations_and_signs_eh\n");
    unit_test_x_c(robotPt,nbtest);
    fprintf(f,"unit_test_x_c\n");
    unit_test_x_e(robotPt,nbtest);
    fprintf(f,"unit_test_x_e\n");
    unit_test_v_c(robotPt,nbtest);
    fprintf(f,"unit_test_v_c\n");
    unit_test_v_e(robotPt,nbtest);
    fprintf(f,"unit_test_v_e\n");
    unit_test_intervals_ac(robotPt,nbtest);
    fprintf(f,"unit_test_intervals_ac\n");
    unit_test_intervals_eh(robotPt,nbtest);
    fprintf(f,"unit_test_intervals_eh\n");
    unit_test_a_b(robotPt,nbtest);
    fprintf(f,"unit_test_a_b\n");
    unit_test_a_g(robotPt,nbtest);
    fprintf(f,"unit_test_a_g\n");
    unit_test_d_x(robotPt,nbtest);
    fprintf(f,"unit_test_d_x\n");
    unit_test_v_optim(robotPt,nbtest);
    fprintf(f,"unit_test_v_optim\n");
    unit_test_synchronize(robotPt,nbtest);
    fprintf(f,"unit_test_synchronize\n");
    unit_test_spline_at_from_q0(robotPt,nbtest);
    unit_test_spline_at_from_qF(robotPt,nbtest);
    unit_test_spline_at(robotPt,nbtest);
    printf("--------------------------------------------------------DONE--\n");
    fclose(f);
}
